/************************************************************************
 *  palmspec.c
 *                      
 *
 ************************************************************************
 *
 * PalmSpec,
 * Spectrum emulator for PalmOS(R)
 *
 * (c)2002 Alex Bassas Serramia
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

#include "const.h"

UInt32 depth = 4;
RGBColorType spalette[16];
int vx=48,vy=16;
UCHAR k=0;

UCHAR cx=0,cy=0;
UCHAR comp = FALSE;
char **list;
int nlist = 0;
UCHAR ocolor,oback;
UCHAR icolor = TRUE;
#ifdef D_SCREEN
UCHAR *sc;
UCHAR m_fc;
UCHAR m_bc;
#endif
UCHAR *memrom;
UCHAR *memram;
DmOpenRef refrom;
DmOpenRef refram;
MemHandle hanrom;
MemHandle hanram;

#ifdef TRACE
void nprintln (void) {
RectangleType rt;

  cy += 10;
  cx = 0;
  if (cy>=150) {
    rt.topLeft.x = 0;
    rt.topLeft.y = 150;
    rt.extent.x = 160;
    rt.extent.y = 10;
    WinEraseRectangle (&rt,0);
    rt.topLeft.x = 0;
    rt.topLeft.y = 0;
    rt.extent.x = 160;
    rt.extent.y = 160;
    WinScrollRectangle (&rt,winUp,10,&rt);
    cy -= 10;
  }
}

void nprint (char *text,long l1,long l2,long l3) {
char cTemp[256];
Int16 l,n;
UCHAR *p=cTemp;

  StrPrintF (cTemp,text,l1,l2,l3);
  for (n=1;n<=StrLen(p);n++) {
    l = FntLineWidth (p,n);
    if (cx+l>=160) {
      WinPaintChars (p,n-1,cx,cy);
      nprintln();
      p += n-1;
      n = 0;
    }
  }
  l = FntLineWidth (p,StrLen(p));
  if (l) {
    WinPaintChars (p,StrLen(p),cx,cy);
    cx += l;
  }
}

UCHAR KbHit (void) {
  return (k);
}

UCHAR GetChar (void) {
UCHAR nk = k;
  k = 0;
  return (nk);
}

ULONG GetTime (void) {
  return (TimGetTicks());
}
#endif

void panic (char *info) {
  ErrDisplay (info);
}

#ifndef D_SCREEN
#define m_WinSetForeColor(c)    WinSetForeColor(c)
#define m_WinSetBackColor(c)    WinSetBackColor(c)
#define m_WinDrawPixel(x,y)     WinDrawPixel(x,y)
#define m_WinErasePixel(x,y)    WinErasePixel(x,y)
#else

void m_putpixel(SHORT x,SHORT y,UCHAR c) {
USHORT i;
CHAR s;

  i = (x>>1) + (y*80);
  s = sc[i];
  if (x&0x01) {
    s = (s&0xf0)|c;
  }
  else {
    s = (s&0x0f)|(c<<4);
  }
  sc[i] = s;
}

#ifndef TRACE
#define m_WinSetForeColor(c)    m_fc=c
#define m_WinSetBackColor(c)    m_bc=c
#define m_WinDrawPixel(x,y)     m_putpixel(x,y,m_fc)
#define m_WinErasePixel(x,y)    m_putpixel(x,y,m_bc)  
#else
#define m_WinSetForeColor(c)    
#define m_WinSetBackColor(c)    
#define m_WinDrawPixel(x,y)     
#define m_WinErasePixel(x,y)    
#endif
#endif

void putpixel (USHORT x,USHORT y,UCHAR color,UCHAR back,UCHAR code,UCHAR mask) {
SHORT yy;

  if (!comp) {
    x -= vx;
    y -= vy;
    if (x>=0 && x<160 && y>=0 && y<160) {
      if (icolor || ocolor!=color) {
        m_WinSetForeColor (color);
        ocolor = color;
      }
      if (icolor || oback!=back) {
        m_WinSetBackColor (back);
        oback = back;
      }
      icolor = FALSE;
      if (mask&0x80) {
        if (code&0x80) m_WinDrawPixel (x+0,y); else m_WinErasePixel (x+0,y);
      }
      if (mask&0x40) {
        if (code&0x40) m_WinDrawPixel (x+1,y); else m_WinErasePixel (x+1,y);
      }
      if (mask&0x20) {
        if (code&0x20) m_WinDrawPixel (x+2,y); else m_WinErasePixel (x+2,y);
      }
      if (mask&0x10) {
        if (code&0x10) m_WinDrawPixel (x+3,y); else m_WinErasePixel (x+3,y);
      }
      if (mask&0x08) {
        if (code&0x08) m_WinDrawPixel (x+4,y); else m_WinErasePixel (x+4,y);
      }
      if (mask&0x04) {
        if (code&0x04) m_WinDrawPixel (x+5,y); else m_WinErasePixel (x+5,y);
      }
      if (mask&0x02) {
        if (code&0x02) m_WinDrawPixel (x+6,y); else m_WinErasePixel (x+6,y);
      }
      if (mask&0x01) {
        if (code&0x01) m_WinDrawPixel (x+7,y); else m_WinErasePixel (x+7,y);
      }
    }
  }
  else {
    yy = (y%8);
    if (yy!=1 && yy!=4 && yy!=7){     
      x = (x/8)*5;
      y = (y/8)*5;
      if (yy == 2) y += 1;
      if (yy == 3) y += 2;
      if (yy == 5) y += 3;
      if (yy == 6) y += 4;
      if (icolor || ocolor!=color) {
        m_WinSetForeColor (color);
        ocolor = color;
      }
      if (icolor || oback!=back) {
        m_WinSetBackColor (back);
        oback = back;
      }
      icolor = FALSE;
      if (mask&0x80) {
        if (code&0x80) m_WinDrawPixel (x+0,y); else m_WinErasePixel (x+0,y);
      }
      if (mask&0x20) {
        if (code&0x20) m_WinDrawPixel (x+1,y); else m_WinErasePixel (x+1,y);
      }
      if (mask&0x10) {
        if (code&0x10) m_WinDrawPixel (x+2,y); else m_WinErasePixel (x+2,y);
      }
      if (mask&0x04) {
        if (code&0x04) m_WinDrawPixel (x+3,y); else m_WinErasePixel (x+3,y);
      }
      if (mask&0x02) {
        if (code&0x02) m_WinDrawPixel (x+4,y); else m_WinErasePixel (x+4,y);
      }
    }
  }
}

void FreeListSNA (void) {
int n;

  if (nlist) {
    for (n=0;n<nlist;n++) {
      MemPtrFree (list[n]);
    }
    MemPtrFree (list);
    nlist = 0;
  }
}

void GetListSNA (void) {
int lFirst;
DmSearchStateType ss;
UInt16 card;
LocalID id;
int n = 0;

  FreeListSNA();
  lFirst = TRUE;
  while (DmGetNextDatabaseByTypeCreator(lFirst,&ss,'Data','PSPe',FALSE,&card,&id)==errNone) {
    lFirst = FALSE;
    n++;
  }
  list = MemPtrNew (sizeof(char *)*n);
  n = 0;
  lFirst = TRUE;
  while (DmGetNextDatabaseByTypeCreator(lFirst,&ss,'Data','PSPe',FALSE,&card,&id)==errNone) {
    lFirst = FALSE;
    list[n] = MemPtrNew (32);
    DmDatabaseInfo (card,id,list[n],NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
    n++;
  }
  nlist = n;
}

void ListSNA (void) {
ListType *lt;
int num,n;

  lt = FrmGetObjectPtr (FrmGetActiveForm(),FrmGetObjectIndex(FrmGetActiveForm(),LIST_SNA));
  if (lt!=NULL) {
    GetListSNA();
    LstSetListChoices (lt,list,nlist);
    LstDrawList (lt);
  }
}

UCHAR Game (void) {
EventType e;
UInt16 error,Exit=FALSE;
UInt16 n;
struct rgb *c;
UInt8 lastkey[255];
UInt8 k;
UInt32 oldkeymask,curstate;
UInt8 flip = 3;

WinHandle winH;
BitmapType *bmpP;

#ifndef TRACE
  WinScreenMode(winScreenModeSet,NULL,NULL,&depth,NULL);
  for (n=0;n<16;n++) {
    c = getrgb (n);
    spalette[n].index = n;
    spalette[n].r = c->r;
    spalette[n].g = c->g;
    spalette[n].b = c->b;
  }
  WinPalette(winPaletteSet,0,16,spalette);
#endif
#ifdef D_SCREEN
  /*
  asm ("move.l #0xfffffa00,%0\n\
        move.l (%0),%0":"=a" (sc):"0" (sc));
  */
  winH = WinGetDisplayWindow();
  bmpP = WinGetBitmap(winH);
  sc = BmpGetBits(bmpP);

#endif
#ifndef TRACE
  Refresh();
#endif
  oldkeymask = KeySetMask (keyBitsAll & 
        ~(keyBitPageUp|keyBitPageDown|keyBitHard1|keyBitHard2|keyBitHard3|keyBitHard4));
  for (n=0;n<256;n++) {
    lastkey[n] = 0;
  }

  while (!Exit) {
    for (n=0;n<256;n++) {
      if (lastkey[n]) {
        lastkey[n]--;
        if (!lastkey[n]) {
          ResetKey (n);
        }
      }
    }
    k = 0;
    curstate = KeyCurrentState();
    if (curstate & keyBitPageUp) k |= KEMPSTON_UP;
    if (curstate & keyBitPageDown) k |= KEMPSTON_DOWN;
    if (curstate & keyBitHard1) k |= KEMPSTON_LEFT;
    if (curstate & keyBitHard2) k |= KEMPSTON_RIGHT;
    if (curstate & keyBitHard3) k |= KEMPSTON_FIRE;
    SetKempston(k);
    if (curstate & keyBitHard4) {
      flip++;
      if (flip>3) flip = 0;
      switch (flip) {
        case 0:
          vx = 0;
          vy = 192-160;
          break;
        case 1:
          vx = 0;
          vy = 0;
          break;
        case 2:
          vx = 256-160;
          vy = 0;
          break;
        case 3:
          vx = 256-160;
          vy = 192-160;
          break;
      }
      Refresh();
    }
    //SetKey ('\n');
    EvtGetEvent (&e, 0);
    if (!SysHandleEvent (&e)) {
      switch (e.eType) {
        case nilEvent:
#ifdef TRACE
          if (!InDebug())
#endif
          MemSemaphoreReserve(true);
          Execute (5000);
          MemSemaphoreRelease(true);
#ifdef TRACE
          if (InDebug()) DoDebug();
#endif
          break;
/*
#ifdef TRACE
        case keyDownEvent:
          k = e.data.keyDown.chr;
          break;
#else 
*/
        case keyDownEvent:
          lastkey[e.data.keyDown.chr] = 20;
          SetKey (e.data.keyDown.chr);
          break;

        case penDownEvent:
          if (!comp) {
            if (e.screenX>=50 && e.screenX<100 &&
                e.screenY>=50 && e.screenY<100) {
              vx = 48;
              vy = 16;
              Refresh();
            }
            else if (e.screenX>=0 && e.screenX<80 &&
                e.screenY>=0 && e.screenY<80) {
              vx = 0;
              vy = 0;
              Refresh();
            }
            else if (e.screenX>80 && e.screenX<160 &&
                e.screenY>=0 && e.screenY<80) {
              vx = 256-160;
              vy = 0;
              Refresh();
            }
            else if (e.screenX>80 && e.screenX<160 &&
                e.screenY>80 && e.screenY<160) {
              vx = 256-160;
              vy = 192-160;
              Refresh();
            }
            else  {
              vx = 0;
              vy = 192-160;
              Refresh();
            }
          }
          break;
        case appStopEvent:
          Exit = TRUE;
          break;
      }
    }
  } 

  KeySetMask (oldkeymask);
  MemSemaphoreReserve(true);
  SaveSNA(TMP_FILE);
  MemSemaphoreRelease(true);
  return (e.eType==appStopEvent);
}


UInt32 PilotMain (UInt16 cmd,void *cmdPBP,UInt16 launchFlags) {
EventType e;
UInt16 error;
UCHAR sel;
UCHAR Exit = FALSE;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    memrom = FileMemAlloc ("tmprom",0x4000,&refrom,&hanrom);
    memram = FileMemAlloc ("tmpram",0xc000,&refram,&hanram);
    MemSemaphoreReserve(true);
    Init();
    MemSemaphoreRelease(true);
    FrmGotoForm (FORM_MAIN);
    do {
      EvtGetEvent (&e,evtWaitForever);
      if (!SysHandleEvent (&e)) {
        if (!MenuHandleEvent (NULL,&e,&error)) {
          switch (e.eType) {
            case frmLoadEvent:
              FrmSetActiveForm (FrmInitForm (e.data.frmLoad.formID));
              break;
            case frmOpenEvent:
              ListSNA();
              FrmDrawForm(FrmGetActiveForm());
              break;
            case ctlEnterEvent:
              switch (e.data.ctlEnter.controlID) {
                case BUTT_LOAD:
                  sel = LstGetSelection(FrmGetObjectPtr (FrmGetActiveForm(),FrmGetObjectIndex(FrmGetActiveForm(),LIST_SNA)));
                  MemSemaphoreReserve(true);
                  LoadSNA(list[sel]);
                  MemSemaphoreRelease(true);
                  Exit = Game();
                  break;
                case BUTT_NEW:
                  Exit = Game();
                  break;
              }
              FrmHandleEvent(FrmGetActiveForm(),&e);
              break;
            default:
              FrmHandleEvent(FrmGetActiveForm(),&e);
              break;
          }
        }
      }
    } 
    while (e.eType!=appStopEvent && !Exit);
    FreeListSNA();
    End();
    FileMemRelease ("tmprom",&refrom,&hanrom);
    FileMemRelease ("tmpram",&refram,&hanram);
  }
  return (errNone);
}
