/************************************************************************
 *  palmstd.c
 *                      
 *
 ************************************************************************
 *
 * PalmSpec,
 * Spectrum emulator for PalmOS(R)
 *
 * (c)2002 Alex Bassas Serramia
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */
#include "const.h"

struct rgb cnormal[8] = { { 0x00,0x00,0x00 },
                                { 0x00,0x00,0xcf },
                                { 0xcf,0x00,0x00 },
                                { 0xcf,0x00,0xcf },
                                { 0x00,0xcf,0x00 },
                                { 0x00,0xcf,0xcf },
                                { 0xcf,0xcf,0x00 },
                                { 0xcf,0xcf,0xcf }  };

struct rgb cbright[8] = { { 0x00,0x00,0x00 },
                                { 0x00,0x00,0xff },
                                { 0xff,0x00,0x00 },
                                { 0xff,0x00,0xff },
                                { 0x00,0xff,0x00 },
                                { 0x00,0xff,0xff },
                                { 0xff,0xff,0x00 },
                                { 0xff,0xff,0xff }  };

#ifdef WIN32
UCHAR FileDump (UCHAR *mem,UCHAR *file,USHORT start,USHORT len) {
UCHAR ret = FALSE;
FILE *f;

  f = fopen(file,"rb");
  if (f!=NULL) {
    fseek (f,start,SEEK_SET);
    fread (mem,1,len,f);
    fclose(f);
    ret = TRUE;
  }
  return (ret);
}

UCHAR FileCreate (UCHAR *file,ULONG size) {
  return (TRUE);
}
  

UCHAR FileWrite (UCHAR *file,UCHAR *mem,ULONG start,ULONG length) {
  return (TRUE);
}

#else

UCHAR FileDump (UCHAR *mem,UCHAR *file,USHORT start,USHORT len) {
UCHAR ret = FALSE;
LocalID id;
DmOpenRef ref;
MemHandle m;
UCHAR *p;
USHORT n;

  id = DmFindDatabase (0,file);
  if (id) {
    ref = DmOpenDatabase (0,id,dmModeReadOnly);
    if (ref) {
      m = DmQueryRecord (ref,0);
      if (m!=NULL) {
        p = (UCHAR *) MemHandleLock (m);
        p += start;
        MemMove (mem,p,len);
        MemHandleUnlock (m);
        ret = TRUE;
      }
      else ErrAlert (DmGetLastErr());
      DmCloseDatabase (ref);
    }
    else ErrAlert (DmGetLastErr());
  }
  else ErrAlert (DmGetLastErr());
  return (ret);
}

UCHAR FileCreate (UCHAR *file,ULONG size) {
LocalID id;
DmOpenRef ref;
MemHandle m;
UInt16 index;

  id = DmFindDatabase(0,file);
  if (id!=0) {
    DmDeleteDatabase(0,id);
  }
  if (DmCreateDatabase (0,file,'PSPe','Data',FALSE)==errNone) {
    id = DmFindDatabase(0,file);
    if (id!=0) {
      ref = DmOpenDatabase (0,id,dmModeReadWrite);
      if (ref) {
        index = 0;
        m = DmNewRecord (ref,&index,size);
        if (m!=0) {
          DmReleaseRecord (ref,0,TRUE);
        }
        else ErrAlert (DmGetLastErr());
        DmCloseDatabase (ref);
      }
      else ErrAlert (DmGetLastErr());
    }
    else ErrAlert (DmGetLastErr());
  }
  else ErrAlert (DmGetLastErr());
  return (TRUE);
}
  

UCHAR FileWrite (UCHAR *file,UCHAR *mem,ULONG start,ULONG len) {
LocalID id;
DmOpenRef ref;
MemHandle m;
UCHAR *p;

  id = DmFindDatabase (0,file);
  if (id) {
    ref = DmOpenDatabase (0,id,dmModeReadWrite);
    if (ref) {
      m = DmGetRecord(ref,0);
      if (m!=NULL) {
        p = (UCHAR *) MemHandleLock (m);
        DmWrite (p,start,mem,len);
        MemHandleUnlock (m);
        DmReleaseRecord (ref,0,FALSE);
      }
      DmCloseDatabase (ref);
    }
  }
  return (TRUE);
}

UCHAR *FileMemAlloc (UCHAR *file,ULONG size,DmOpenRef *ref,MemHandle *m) {
LocalID id;
UInt16 index;
UCHAR *p;

  id = DmFindDatabase(0,file);
  if (id!=0) {
    DmDeleteDatabase(0,id);
  }
  if (DmCreateDatabase (0,file,'PSPe','Temp',FALSE)==errNone) {
    id = DmFindDatabase(0,file);
    if (id!=0) {
      *ref = DmOpenDatabase (0,id,dmModeReadWrite);
      if (*ref) {
        index = 0;
        *m = DmNewRecord (*ref,&index,size);
        if (*m!=0) {
          p = (UCHAR *) MemHandleLock (*m);
          return (p);
        }
        else ErrAlert (DmGetLastErr());
        DmCloseDatabase (*ref);
      }
      else ErrAlert (DmGetLastErr());
    }
    else ErrAlert (DmGetLastErr());
  }
  else ErrAlert (DmGetLastErr());
  return (NULL);
}

void FileMemRelease (UCHAR *file,DmOpenRef *ref,MemHandle *m) {
LocalID id;

  MemHandleUnlock (*m);
  DmReleaseRecord (*ref,0,FALSE);
  DmCloseDatabase (*ref);
  id = DmFindDatabase(0,file);
  if (id!=0) {
    DmDeleteDatabase(0,id);
  }
}

#endif

#ifdef WIN32

void *MemPtrNew (ULONG size) {
  return (malloc(size));
}

UCHAR MemSet (void *dest,UINT bytes,INT value) {
  memset (dest,value,bytes);
  return (0);
}

UCHAR MemPtrFree (void *p) {
  free (p);
  return (0);
}

#else

void *memptrnew (ULONG size) {
  return (MemPtrNew(size));
}

UCHAR memptrfree (void *p) {
  return (MemPtrFree(p));
}

#endif

struct rgb *getrgb (UCHAR color) {
  return ((color>7)?&cbright[color-8]:&cnormal[color]);
}