/************************************************************************
 *  spectrum.h
 *                      
 *
 ************************************************************************
 *
 * PalmSpec,
 * Spectrum emulator for PalmOS(R)
 *
 * (c)2002 Alex Bassas Serramia
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */
#ifndef _spectrum_h
#define _spectrum_h

#pragma pack(1)

#define S_FLAG    0x80
#define Z_FLAG    0x40
#define X1_FLAG   0x20
#define H_FLAG    0x10
#define X2_FLAG   0x08
#define P_FLAG    0x04
#define N_FLAG    0x02
#define C_FLAG    0x01

#define SCREEN_I  0x4000
#define SCREEN_F  0x57ff
#define COLOR_I   0x5800
#define COLOR_F   0x5aff

#define SCREEN_BX 32
#define SCREEN_BY 192

#define COLOR_BX  32
#define COLOR_BY  24

#define M_FLASH   0x80
#define M_BRIGHT  0x40
#define M_B_COLOR 0x38
#define M_F_COLOR 0x07

#define KEMPSTON_PORT   0x1f
#define KEMPSTON_FIRE   0x10
#define KEMPSTON_UP     0x08
#define KEMPSTON_DOWN   0x04
#define KEMPSTON_LEFT   0x02
#define KEMPSTON_RIGHT  0x01

struct rgb {
  UCHAR r;
  UCHAR g;
  UCHAR b;
};

#define ROM_FILE "SpectrumRom"
#define TMP_FILE "LastSession"

#ifdef TRACE
void print (char *text,long l1,long l2,long l3);
void println (void);
void nprint (char *text,long l1,long l2,long l3);
void nprintln (void);
UCHAR InDebug (void);
void DoDebug (void);
ULONG GetTime(void);
#else
#define print(a,b,c,d)
#define println()
#define InDebug() FALSE
#define DoDebug() 
#endif
void SetKey (UCHAR k);
void ResetKey (UCHAR k);
void SetKempston (UCHAR k);
void panic (char *info);
void Init (void);
void End (void);
void Execute (ULONG time);
void putpixel (USHORT x,USHORT y,UCHAR bc,UCHAR bb,UCHAR code,UCHAR mask);
void Refresh (void);
UCHAR LoadSNA (char *file);
UCHAR SaveSNA (char *file);
UCHAR KbHit (void) ;
UCHAR GetChar (void) ;

#endif  //_spectrum_h