/************************************************************************
 *  spectrums.h
 *                      
 *
 ************************************************************************
 *
 * PalmSpec,
 * Spectrum emulator for PalmOS(R)
 *
 * (c)2002 Alex Bassas Serramia
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */
 /*

	%d0 = ...
	%d1 = BC		AF
	%d2 = DE		HL
	%d3 = IR		INCR/FLAGS: XNZVC
	%d4 = SP		PC
	%d5 = IX		IY
	%d6 = time
	%d7 = ...


	%a0 = ...
	%a1 = ...
	%a2 = return addr
	%a3 = rom
	%a4 = mem
	%a5 = globals
	%a6 = link
	%a7 = stack

*/
  
# -------------------------------------------------------------------------------------- 

.text
			
.macro S_A
	ror.w #8,%d1
	.endm

.macro E_A
	ror.w #8,%d1
	.endm

.macro S_F
	.endm

.macro E_F
	.endm

.macro S_AF
	.endm

.macro E_AF
	.endm

.macro S_B		
	swap %d1
	ror.w #8,%d1
	.endm

.macro E_B		
	ror.w #8,%d1
	swap %d1
	.endm

.macro S_C
	swap %d1
	.endm

.macro E_C
	swap %d1	
	.endm

.macro S_BC
	swap %d1
	.endm

.macro E_BC
	swap %d1
	.endm

.macro S_D
	swap %d2
	ror.w #8,%d2
	.endm

.macro E_D
	ror.w #8,%d2
	swap %d2
	.endm

.macro S_E
	swap %d2
	.endm

.macro E_E	
	swap %d2	
	.endm
		
.macro S_DE
	swap %d2
	.endm
	
.macro E_DE
	swap %d2
	.endm
		
.macro S_H
	ror.w #8,%d2
	.endm

.macro E_H		
	ror.w #8,%d2
	.endm

.macro S_L
	.endm

.macro E_L		
	.endm

.macro S_HL
	.endm

.macro E_HL
	.endm

.macro S_INCR
	ror.w #8,%d3
	.endm

.macro E_INCR
	ror.w #8,%d3
	.endm

.macro S_FLAGS
	.endm

.macro E_FLAGS
	.endm

.macro S_IR
	swap %d3
	.endm

.macro E_IR
	swap %d3
	.endm

.macro S_time
	.endm

.macro E_time
	.endm

.macro S_SP_S
	swap %d4
	ror.w #8,%d4
	.endm

.macro E_SP_S
	ror.w #8,%d4
	swap %d4
	.endm

.macro S_SP_P
	swap %d4
	.endm

.macro E_SP_P
	swap %d4
	.endm

.macro S_SP
	swap %d4
	.endm

.macro E_SP
	swap %d4
	.endm

.macro S_PC_P
	ror.w #8,%d4
	.endm

.macro E_PC_P
	ror.w #8,%d4
	.endm

.macro S_PC_C
	.endm

.macro E_PC_C
	.endm

.macro S_PC
	.endm

.macro E_PC
	.endm

.macro S_I
	swap %d3
	ror.w #8,%d3
	.endm

.macro E_I
	ror.w #8,%d3
	swap %d3
	.endm

.macro S_R		
	swap %d3
	.endm

.macro E_R	
	swap %d3	
	.endm

.macro S_IX_I
	swap %d5
	ror.w #8,%d5
	.endm

.macro E_IX_I
	ror.w #8,%d5
	swap %d5
	.endm

.macro S_IX_X
	swap %d5
	.endm

.macro E_IX_X
	swap %d5
	.endm

.macro S_IX
	swap %d5
	.endm

.macro E_IX
	swap %d5
	.endm

.macro S_IY_I
	ror.w #8,%d5
	.endm

.macro E_IY_I
	ror.w #8,%d5
	.endm

.macro S_IY_Y
	.endm

.macro E_IY_Y
	.endm

.macro S_IY		
	.endm

.macro E_IY		
	.endm

.macro GETB var,reg
	lea \var@END.w(%a5),%a0
	move.b (%a0),\reg
	.endm	

.macro SETB var,reg
	lea \var@END.w(%a5),%a0
	move.b \reg,(%a0)
	.endm

.macro ANDB var,reg
	lea \var@END.w(%a5),%a0
	and.b \reg,(%a0)
	.endm

.macro ORB var,reg
	lea \var@END.w(%a5),%a0
	or.b \reg,(%a0)
	.endm

.macro GETW var,reg
	lea \var@END.w(%a5),%a0
	move.w (%a0),\reg
	.endm	

.macro SETW var,reg
	lea \var@END.w(%a5),%a0
	move.w \reg,(%a0)
	.endm

.macro GETL var,reg
	lea \var@END.w(%a5),%a0
	move.l (%a0),\reg
	.endm	

.macro SETL var,reg
	lea \var@END.w(%a5),%a0
	move.l \reg,(%a0)
	.endm

.macro SUML var,reg
	lea \var@END.w(%a5),%a0
	add.l \reg,(%a0)
	.endm

.macro RESL var,reg
	lea \var@END.w(%a5),%a0
	sub.l \reg,(%a0)
	.endm

.macro BREAK
	trap #8
	.endm

.macro PUSHALL
	move.l %d0,-(%sp)
	move.l %d1,-(%sp)
	move.l %d2,-(%sp)
	move.l %d3,-(%sp)
	move.l %d4,-(%sp)
	move.l %d5,-(%sp)
	move.l %d6,-(%sp)
	move.l %d7,-(%sp)
	move.l %a0,-(%sp)
	move.l %a1,-(%sp)
	move.l %a2,-(%sp)
	move.l %a3,-(%sp)
	move.l %a4,-(%sp)
	move.l %a5,-(%sp)
	move.l %a6,-(%sp)
	.endm

.macro POPALL
	move.l (%sp)+,%a6
	move.l (%sp)+,%a5
	move.l (%sp)+,%a4
	move.l (%sp)+,%a3
	move.l (%sp)+,%a2
	move.l (%sp)+,%a1
	move.l (%sp)+,%a0
	move.l (%sp)+,%d7
	move.l (%sp)+,%d6
	move.l (%sp)+,%d5
	move.l (%sp)+,%d4
	move.l (%sp)+,%d3
	move.l (%sp)+,%d2
	move.l (%sp)+,%d1
	move.l (%sp)+,%d0
	.endm

.macro PRINT s,a,b,c
	#void nprint (char *text,long l1,long l2,long l3)
	move.l \c,-(%sp)
	move.l \b,-(%sp)
	move.l \a,-(%sp)
	lea \s(%pc),%a0 
	move.l %a0,-(%sp)
	lea nprint(%pc),%a0
	jsr (%a0)
	add.l #16,%sp
	.endm

.macro PRINTLN
	#void nprintln (void);
	lea nprintln(%pc),%a0
	jsr (%a0)
	.endm

.macro PUTPIXEL x,y,bc,bb,code,mask
	#void putpixel (USHORT x,USHORT y,UCHAR bc,UCHAR bb,UCHAR code,UCHAR mask)
	move.b \mask,-(%sp)
	move.b \code,-(%sp)
	move.b \bb,-(%sp)
	move.b \bc,-(%sp)
	move.w \y,-(%sp)
	move.w \x,-(%sp)
	lea putpixel(%pc),%a0
	jsr (%a0)
	add.l #12,%sp
	.endm

.macro LOAD_R r,reg
	S_\r
	GETB REG_\r,\reg
	E_\r
	.endm

.macro LOAD_W r,reg
	S_\r
	GETW REG_\r,\reg
	E_\r
	.endm

.macro RESET_REGS
	SETB REG_A,#0
	SETB REG_F,#0
	SETB REG_B,#0
	SETB REG_C,#0
	SETB REG_D,#0
	SETB REG_E,#0
	SETB REG_H,#0
	SETB REG_L,#0
	SETB REG_I,#0
	SETB REG_R,#0

	SETB REG_FLAGS,#0

	SETB REG_IM,#0
	SETB REG_IFF1,#0
	SETB REG_IFF2,#0
	SETB REG_NEI,#0
	SETB REG_flash,#0

	SETW REG_SP,#0
	SETW REG_PC,#0
	SETW REG_IX,#0
	SETW REG_IY,#0

	SETB KEYBOARD0,#0xff
	SETB KEYBOARD1,#0xff
	SETB KEYBOARD2,#0xff
	SETB KEYBOARD3,#0xff
	SETB KEYBOARD4,#0xff
	SETB KEYBOARD5,#0xff
	SETB KEYBOARD6,#0xff
	SETB KEYBOARD7,#0xff

	SETB KEMPSTON,#0x00
	
	SETL time,#0
	SETL TIME_INT,#0
	SETL TIME_FLASH,#0
	
	SETB trace,#0

	SETB rsc,#0
	.endm

.macro LOAD_REGS
	LOAD_R A,%d1
	LOAD_R F,%d1
	LOAD_R B,%d1
	LOAD_R C,%d1
	LOAD_R D,%d2
	LOAD_R E,%d2
	LOAD_R H,%d2
	LOAD_R L,%d2
	LOAD_R I,%d3
	LOAD_R R,%d3
	LOAD_R FLAGS,%d3
	LOAD_W SP,%d4
	LOAD_W PC,%d4
	LOAD_W IX,%d5
	LOAD_W IY,%d5
	GETL ROM,%a3
	GETL MEM,%a4
	GETL time,%d6
	F_TO_FLAGS
	.endm

.macro SAVE_R r,reg
	S_\r
	SETB REG_\r,\reg
	E_\r
	.endm

.macro SAVE_W r,reg
	S_\r
	SETW REG_\r,\reg
	E_\r
	.endm

.macro SAVE_REGS
	FLAGS_TO_F
	SAVE_R A,%d1
	SAVE_R F,%d1
	SAVE_R B,%d1
	SAVE_R C,%d1
	SAVE_R D,%d2
	SAVE_R E,%d2
	SAVE_R H,%d2
	SAVE_R L,%d2
	SAVE_R I,%d3
	SAVE_R R,%d3
	SAVE_R FLAGS,%d3
	SAVE_W SP,%d4
	SAVE_W PC,%d4
	SAVE_W IX,%d5
	SAVE_W IY,%d5
	SETL time,%d6
	.endm

.macro LDB	r,s,reg
	S_\r
	move.b \s,\reg
	E_\r
	.endm

.macro STB	r,s,reg
	S_\r
	move.b \reg,\s
	E_\r
	.endm

.macro LDW	r,s,reg
	S_\r
	move.w \s,\reg
	E_\r
	.endm

.macro STW	r,s,reg
	S_\r
	move.w \reg,\s
	E_\r
	.endm

.macro ADDB r,s,reg
	S_\r
	add.b \reg,\s
	E_\r
	.endm

.macro SUBB r,s,reg
	S_\r
	sub.b \reg,\s
	E_\r
	.endm

.macro ADDW r,s,reg
	S_\r
	add.w \reg,\s
	E_\r
	.endm

.macro SUBW r,s,reg
	S_\r
	sub.w \reg,\s
	E_\r
	.endm

.macro ADDL r,s,reg
	S_\r
	add.l \reg,\s
	E_\r
	.endm

.macro STFLAGS reg			# Save 68k flags.
	move %sr,\reg
	move.b \reg,%d3
	.endm

.macro STFLAGSC reg			# Save 68k flags without carry.
	move %sr,\reg
	btst.b #0,%d3
	bne stf0\@
	bclr.b #0,\reg
	bra stf1\@
	.even
stf0\@:
	bset.b #0,\reg
stf1\@:
	move.b \reg,%d3
	.endm

.macro STFLAGSCP reg		# Save 68k flags without carry and parity.
	move %sr,\reg
	btst.b #0,%d3
	bne stf0\@
	bclr.b #0,\reg
	bra stf1\@
	.even
stf0\@:
	bset.b #0,\reg
stf1\@:
	btst.b #1,%d3
	bne stf2\@
	bclr.b #1,\reg
	bra stf3\@
	.even
stf2\@:
	bset.b #1,\reg
stf3\@:
	move.b \reg,%d3
	.endm

.macro STFLAGSOC 			# Save 68k carry.
	bcc stfo0\@
	bset.b #0,%d3
	bra stfo1\@
	.even
stfo0\@:
	bclr.b #0,%d3
stfo1\@:
	.endm

.macro RTFLAGS				# Restore 68k flags.
	move %d3,%ccr
	.endm

.macro MAP_BIT src,sb,dst,db
	btst.b #\sb,\src
	beq map0\@
	bset.b #\db,\dst
	bra map1\@
	.even
map0\@:
	bclr.b #\db,\dst
map1\@:
	.endm

/*
       7  6  5  4  3  2  1  0

68k  = _  _  _  X  N  Z  V  C

z80a = S  Z  X1 H  X2 P  N  C

*/

.macro F_TO_FLAGS
	MAP_BIT %d1,0,%d3,0
	MAP_BIT %d1,2,%d3,1
	MAP_BIT %d1,6,%d3,2
	MAP_BIT %d1,7,%d3,3
	.endm

.macro FLAGS_TO_F
	MAP_BIT %d3,0,%d1,0
	MAP_BIT %d3,1,%d1,2
	MAP_BIT %d3,2,%d1,6
	MAP_BIT %d3,3,%d1,7
	.endm

.macro MEM_ALLOC l,r
	#void *memptrnew (ULONG size);
	move.l #\l,-(%sp)
	bsr.w memptrnew
	move.l %a0,\r
	add.l #4,%sp
	.endm

.macro MEM_FREE r
	#UCHAR memptrfree (void *p);
	move.l \r,-(%sp)
	bsr.w memptrfree
	add.l #4,%sp
	.endm

.macro FILEDUMP r,name,init,len
	#UCHAR FileDump(UCHAR *mem,UCHAR *file,USHORT start,USHORT len)
	move.w #\len,-(%sp)
	move.w #\init,-(%sp)
	move.l \name,-(%sp)
	move.l \r,-(%sp)
	bsr.w FileDump
	add.l #12,%sp
	.endm

.macro FILECREATE name,len
	#UCHAR FileCreate (UCHAR *file,ULONG size)
	move.l #\len,-(%sp)
	move.l \name,-(%sp)
	bsr.w FileCreate
	add.l #8,%sp
	.endm

.macro FILEWRITE name,mem,start,len
	#UCHAR FileWrite (UCHAR *file,UCHAR *mem,ULONG start,ULONG length)
	move.l #\len,-(%sp)
	move.l #\start,-(%sp)
	move.l \mem,-(%sp)
	move.l \name,-(%sp)
	bsr.w FileWrite
	add.l #16,%sp
	.endm

.macro RET ip,t
	ADD_PC #\ip
	ADD_time #\t
	bra nextins
	.endm

.macro IRET i
	ADD_time #\i
	.endm

.macro SRET i
	ADD_PC #\i
	.endm

.macro RET0
	bra nextins
	.endm

.macro LOAD_OPCODE	opcode,defunc
	lea \opcode@END.w(%a5),%a0
	move.w #256,%d0
	lea \defunc(%pc),%a1
next\@:
	move.l %a1,(%a0)
	add.l #4,%a0
	sub.w #1,%d0
	bne next\@
	.endm

.macro SET_OPCODE	num,func
	lea \func(%pc),%a1
	move.l %a1,\num*4(%a0)
	.endm

.macro EXEC_CODE	opcode,code
	lea \opcode@END.w(%a5),%a0
	and.w #0xff,\code
	add.w \code,\code
	add.w \code,\code
	move.l (%a0,\code.w),%a0
	jmp (%a0)
	.endm

.macro FUNC s
	.even
\s:
	.endm

.macro READB0
	# %a0 = z80dir
	# exit: %d0 = code
	lea readb0\@(%pc),%a2
	bra readb0
	.even
readb0\@:
	.endm


.macro READB
	# %a0 = z80dir
	# %d0 = desp
	# exit: %d0 = code
	lea readb\@(%pc),%a2
	bra readb
	.even
readb\@:
	.endm

.macro READW0
	# %a0 = z80 dir
	# exit: %d0 = code
	lea readw0\@(%pc),%a2
	bra readw0
	.even
readw0\@:
	.endm

.macro READW
	# %a0 = z80 dir
	# %d0 = desp
	# exit: %d0 = code
	lea readw\@(%pc),%a2
	bra readw
	.even
readw\@:
	.endm

.macro STORE
	# %a0 = z80 dir
	# %d0 = code
	lea store\@(%pc),%a2
	bra store0
	.even
store\@:
	.endm

.macro STOREW
	# %a0 = z80 dir
	# %d0 = code
	exg %a0,%d0
	cmp.w #0x4000,%d0
	bcs store5\@
	sub.w #0x4000,%d0
	and.l #0xffff,%d0
	exg %a0,%d0
	move.b %d0,(%a4,%a0.l)
	exg %a0,%d0
	add.w #1,%d0
	beq store5\@
	exg %a0,%d0
	ror.w #8,%d0
	move.b %d0,(%a4,%a0.l)
	bra store1\@
	.even
store5\@:
	exg %a0,%d0
store1\@:
	.endm

.macro PCONTB0 reg
	GET_\reg %a0
	READB0
	.endm

.macro PCONTB reg,desp
	GET_\reg %a0
	move.b \desp,%d0
	READB
	.endm

.macro PCONTBR0 reg
	move.w \reg %a0
	READB0
	.endm

.macro PCONTBR reg,desp
	move.w \reg %a0
	move.b \desp,%d0
	READB
	.endm

.macro PCONTW0 reg
	GET_\reg %a0
	READW0
	.endm

.macro PCONTW reg,desp
	GET_\reg %a0
	move.b \desp,%d0
	READW
	.endm

.macro PCONTWR0 reg
	move.w \reg,%a0
	READW0
	.endm

.macro PCONTWR reg,desp
	move.w \reg,%a0
	move.w \desp,%d0
	READW
	.endm

.macro GET_A reg
	LDB A,%d1,\reg
	.endm

.macro GET_F reg
	LDB F,%d1,\reg
	.endm

.macro GET_B reg
	LDB B,%d1,\reg
	.endm

.macro GET_C reg
	LDB C,%d1,\reg
	.endm

.macro GET_D reg
	LDB D,%d2,\reg
	.endm

.macro GET_E reg
	LDB E,%d2,\reg
	.endm

.macro GET_H reg
	LDB H,%d2,\reg
	.endm

.macro GET_L reg
	LDB L,%d2,\reg
	.endm

.macro GET_I reg
	LDB I,%d3,\reg
	.endm

.macro GET_R reg
	LDB R,%d3,\reg
	.endm

.macro GET_INCR reg
	LDB INCR,%d3,\reg
	.endm

.macro GET_FLAGS reg
	LDB FLAGS,%d3,\reg
	.endm

.macro GET_IM reg
	GETB REG_IM,\reg
	.endm

.macro GET_IFF1 reg
	GETB REG_IFF1,\reg
	.endm

.macro GET_IFF2 reg
	GETB REG_IFF2,\reg
	.endm

.macro GET_NEI reg
	GETB REG_NEI,\reg
	.endm

.macro GET_flash reg
	GETB REG_flash,\reg
	.endm

.macro GET_AF reg
	LDW AF,%d1,\reg
	.endm

.macro GET_BC reg
	LDW BC,%d1,\reg
	.endm

.macro GET_DE reg
	LDW DE,%d2,\reg
	.endm

.macro GET_HL reg
	LDW HL,%d2,\reg
	.endm

.macro GET_IR reg
	LDW IR,%d3,\reg
	.endm

.macro GET_IX reg
	LDW IX,%d5,\reg
	.endm

.macro GET_IY reg
	LDW IY,%d5,\reg
	.endm

.macro GET_SP reg
	LDW SP,%d4,\reg
	.endm

.macro GET_PC reg
	LDW PC,%d4,\reg
	.endm

.macro GET_PC_P reg
	LDB PC_P,%d4,\reg
	.endm

.macro GET_PC_C reg
	LDB PC_C,%d4,\reg
	.endm

.macro GET_SP_S reg
	LDB SP_S,%d4,\reg
	.endm

.macro GET_SP_P reg
	LDB SP_P,%d4,\reg
	.endm

.macro GET_IX_I reg
	LDB IX_I,%d5,\reg
	.endm

.macro GET_IX_X reg
	LDB IX_X,%d5,\reg
	.endm

.macro GET_IY_I reg
	LDB IY_I,%d5,\reg
	.endm

.macro GET_IY_Y reg
	LDB IY_Y,%d5,\reg
	.endm

.macro GET_time reg
	move.l %d6,\reg
	.endm

.macro SET_A reg
	STB A,%d1,\reg
	.endm

.macro SET_F reg
	STB F,%d1,\reg
	.endm

.macro SET_B reg
	STB B,%d1,\reg
	.endm

.macro SET_C reg
	STB C,%d1,\reg
	.endm

.macro SET_D reg
	STB D,%d2,\reg
	.endm

.macro SET_E reg
	STB E,%d2,\reg
	.endm

.macro SET_H reg
	STB H,%d2,\reg
	.endm

.macro SET_L reg
	STB L,%d2,\reg
	.endm

.macro SET_I reg
	STB I,%d3,\reg
	.endm

.macro SET_R reg
	STB R,%d3,\reg
	.endm

.macro SET_INCR reg
	STB INCR,%d3,\reg
	.endm

.macro SET_FLAGS
	STB FLAGS,%d3,\reg
	.endm

.macro SET_IM reg
	SETB REG_IM,\reg
	.endm

.macro SET_IFF1 reg
	SETB REG_IFF1,\reg
	.endm

.macro SET_IFF2 reg
	SETB REG_IFF2,\reg
	.endm

.macro SET_NEI reg
	SETB REG_NEI,\reg
	.endm

.macro SET_flash reg
	SETB REG_flash,\reg
	.endm

.macro SET_AF reg
	STW AF,%d1,\reg
	.endm

.macro SET_BC reg
	STW BC,%d1,\reg
	.endm

.macro SET_DE reg
	STW DE,%d2,\reg
	.endm

.macro SET_HL reg
	STW HL,%d2,\reg
	.endm

.macro SET_IX reg
	STW IX,%d5,\reg
	.endm

.macro SET_IY reg
	STW IY,%d5,\reg
	.endm

.macro SET_SP reg
	STW SP,%d4,\reg
	.endm

.macro SET_PC reg
	STW PC,%d4,\reg
	.endm

.macro SET_PC_P reg
	STB PC_P,%d4,\reg
	.endm

.macro SET_PC_C reg
	STB PC_C,%d4,\reg
	.endm

.macro SET_SP_S reg
	STB SP_S,%d4,\reg
	.endm

.macro SET_SP_P reg
	STB SP_P,%d4,\reg
	.endm

.macro SET_IX_I reg
	STB IX_I,%d5,\reg
	.endm

.macro SET_IX_X reg
	STB IX_X,%d5,\reg
	.endm

.macro SET_IY_I reg
	STB IY_I,%d5,\reg
	.endm

.macro SET_IY_Y reg
	STB IY_Y,%d5,\reg
	.endm

.macro SET_time reg
	move.l \reg,%d6
	.endm

.macro SET_FLAGS reg
	STB FLAGS,%d3,\reg
	.endm

.macro ADD_R reg
	ADDB R,%d3,\reg
	.endm

.macro ADD_BC reg
	ADDW BC,%d1,\reg
	.endm

.macro ADD_DE reg
	ADDW DE,%d2,\reg
	.endm

.macro ADD_HL reg
	ADDW HL,%d2,\reg
	.endm

.macro ADD_IX reg
	ADDW IX,%d5,\reg
	.endm

.macro ADD_IY reg
	ADDW IY,%d5,\reg
	.endm

.macro ADD_PC reg
	ADDW PC,%d4,\reg
	.endm

.macro ADD_SP reg
	ADDW SP,%d4,\reg
	.endm

.macro ADD_time reg
	ADDL time,%d6,\reg
	.endm

.macro SUB_B reg
	SUBB B,%d1,\reg
	.endm

.macro SUB_BC reg
	SUBW BC,%d1,\reg
	.endm

.macro SUB_DE reg
	SUBW DE,%d2,\reg
	.endm

.macro SUB_HL reg
	SUBW HL,%d2,\reg
	.endm

.macro SUB_IX reg
	SUBW IX,%d5,\reg
	.endm

.macro SUB_IY reg
	SUBW IY,%d5,\reg
	.endm

.macro SUB_SP reg
	SUBW SP,%d4,\reg
	.endm

.macro FLAG0_C
	bclr.b #0,%d3
	.endm

.macro FLAG0_N
	bclr.b #1,%d1
	.endm

.macro FLAG0_P
	bclr.b #1,%d3
	.endm

.macro FLAG0_H
	bclr.b #4,%d1
	.endm

.macro FLAG0_Z
	bclr.b #2,%d3
	.endm

.macro FLAG0_S
	bclr.b #3,%d3
	.endm

.macro FLAG1_C
	bset.b #0,%d3
	.endm

.macro FLAG1_N
	bset.b #1,%d1
	.endm

.macro FLAG1_P
	bset.b #1,%d3
	.endm

.macro FLAG1_H
	bset.b #4,%d1
	.endm

.macro FLAG1_Z
	bset.b #2,%d3
	.endm

.macro FLAG1_S
	bset.b #3,%d3
	.endm

.macro FLAGC_C
	bchg.b #0,%d3
	.endm

.macro FLAG_Z l
	btst.b #2,%d3
	bne \l
	.endm

.macro FLAG_NZ l
	btst.b #2,%d3
	beq \l
	.endm

.macro FLAG_C l
	btst.b #0,%d3
	bne \l
	.endm

.macro FLAG_NC l
	btst.b #0,%d3
	beq \l
	.endm

.macro FLAG_PE l
	btst.b #1,%d3
	bne \l
	.endm

.macro FLAG_PO l
	btst.b #1,%d3
	beq \l
	.endm

.macro FLAG_M l
	btst.b #3,%d3
	bne \l
	.endm

.macro FLAG_P l
	btst.b #3,%d3
	beq \l
	.endm

.macro FLAG_H l
	btst.b #4,%d1
	bne \l
	.endm

.macro FLAG_NH l
	btst.b #4,%d1
	beq \l
	.endm

.macro TESTFN f,r,b
	btst.b #\b,\r
	bne testfa\@
	FLAG1_\f
	bra testfb\@
testfa\@:
	FLAG0_\f
testfb\@:
	.endm

.macro SETCX reg
	GET_FLAGS \reg
	asr.b #1,\reg
	.endm

.macro ADDFLAG_H
	# H
	move.w %d0,%a0
	move.w %d7,%a1
	and.b #0x0f,%d0
	and.b #0x0f,%d7 
	add.b %d7,%d0
	and.b #0xf0,%d0
	#cmp.b #0,%d0
	bne addhp\@
	FLAG0_H
	bra addh\@
	.even
addhp\@:
	FLAG1_H
addh\@:
	move.w %a0,%d0
	move.w %a1,%d7
.endm

.macro ADDHFLAG_H
	# H
	move.w %d0,%a0
	move.w %d7,%a1
	and.w #0x0f00,%d0
	and.w #0x0f00,%d7 
	add.w %d7,%d0
	and.w #0xf000,%d0
	#cmp.w #0,%d6
	bne addhp\@
	FLAG0_H
	bra addh\@
	.even
addhp\@:
	FLAG1_H
addh\@:
	move.w %a0,%d0
	move.w %a1,%d7
.endm

.macro SUBFLAG_H
	# H
	move.w %d0,%a0
	move.w %d7,%a1
	and.b #0xf0,%d0
	and.b #0xf0,%d7 
	cmp.b %d0,%d7
	beq addhz\@
	bcc addhp\@
addhz\@:
	FLAG0_H
	bra addh\@
	.even
addhp\@:
	FLAG1_H
addh\@:
	move.w %a0,%d0
	move.w %a1,%d7
.endm

.macro SUBHFLAG_H
	# H
	move.w %d0,%a0
	move.w %d7,%a1
	and.w #0xf000,%d0
	and.w #0xf000,%d7 
	cmp.w %d0,%d7
	beq addhz\@
	bcc addhp\@
addhz\@:
	FLAG0_H
	bra addh\@
	.even
addhp\@:
	FLAG1_H
addh\@:
	move.w %a0,%d0
	move.w %a1,%d7
.endm

.macro ADDFLAGSC
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra addflagsc
	.even
add\@:
	.endm

.macro ADDFLAGS 
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra addflags
	.even
add\@:
	.endm

.macro ADDHFLAGSC
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra addhflagsc
	.even
add\@:
	.endm

.macro ADDHFLAGS
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra addhflags
	.even
add\@:
	.endm

.macro SUBFLAGSC
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra subflagsc
	.even
add\@:
	.endm

.macro SUBFLAGS
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra subflags
	.even
add\@:
	.endm

.macro SUBHFLAGS
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	lea add\@(%pc),%a2
	bra subhflags
	.even
add\@:
	.endm

.macro XORFLAGS
	# %d0 = result
	STFLAGS %d7
	FLAG0_C
	FLAG0_N
	TESTFN P,%d0,0
	FLAG0_H
	.endm

.macro ANDFLAGS 
	# %d0 = result
	STFLAGS %d7
	FLAG0_C
	FLAG0_N
	TESTFN P,%d0,0
	FLAG1_H
	.endm

.macro R_FLAGSC
	# %d0 = result
	STFLAGS %d7
	TESTFN P,%d0,0
	FLAG0_H
	FLAG0_N
	.endm

.macro OUTPORT w,c
	.endm

.macro INPORT w
	# %d0 = port
	# %d0 = ret data
	move.w \w,%d0
	lea port\@(%pc),%a2
	bra inport
	.even
port\@:
	.endm

# --------------------------------------------------------------------------------------

FUNC noinstr
	BREAK
	RET 0,1

FUNC NOP
	SET_INCR #1
	RET 1,1

FUNC NOPNOP
	RET 2,2

FUNC LD_SP_HL
	S_HL
	S_SP
	move.w %d2,%d4
	E_SP
	E_HL
	RET 1,5

FUNC PUSH_AF
	FLAGS_TO_F
	GET_AF %d0
	S_SP
	sub.w #2,%d4
	move.w %d4,%a0
	STOREW
	E_SP
	RET 1,3


.macro F_PUSH h,l,r
FUNC PUSH_\h\l
	GET_\h\l %d0
	S_SP
	sub.w #2,%d4
	move.w %d4,%a0
	STOREW
	E_SP
	RET 1,3
	.endm

F_PUSH B,C,%d1
F_PUSH D,E,%d2
F_PUSH H,L,%d2

FUNC POP_AF
	S_SP
	S_AF
	PCONTWR0 %d4
	move.w %d0,%d1
	add.w #2,%d4
	E_AF
	E_SP
	F_TO_FLAGS
	RET 1,3

.macro F_POP h,l,r
FUNC POP_\h\l
	S_SP
	S_\h\l
	PCONTWR0 %d4
	move.w %d0,\r
	add.w #2,%d4
	E_\h\l
	E_SP
	RET 1,3
	.endm

F_POP B,C,%d1
F_POP D,E,%d2
F_POP H,L,%d2

FUNC ADD_A_HL_
	PCONTB0 HL
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 1,2

	.even
f_add_a1:
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 1,1

.macro F_ADD_A l
FUNC ADD_A_\l
	GET_\l %d0
	bra f_add_a1
	.endm

F_ADD_A B
F_ADD_A C
F_ADD_A D
F_ADD_A E
F_ADD_A H
F_ADD_A L
F_ADD_A A

FUNC ADD_A_n
	PCONTB PC,#1
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 2,2

FUNC XOR_HL_
	PCONTB0 HL
	GET_A %d7
	exg %d0,%d7
	eor.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 1,2

	.even
f_xor1:
	GET_A %d0
	eor.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 1,2

.macro F_XOR l
FUNC XOR_\l
	GET_\l %d7
	bra f_xor1
	.endm

F_XOR B
F_XOR C
F_XOR D
F_XOR E
F_XOR H
F_XOR L
F_XOR A

FUNC XOR_n
	PCONTB PC,#1
	GET_A %d7
	exg %d0,%d7
	eor.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 2,2

FUNC ADC_A_HL_
	PCONTB0 HL
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 1,2

	.even
f_adc_a1:
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 1,2

.macro F_ADC_A l
FUNC ADC_A_\l
	GET_\l %d0
	bra f_adc_a1
	.endm

F_ADC_A B
F_ADC_A C
F_ADC_A D
F_ADC_A E
F_ADC_A H
F_ADC_A L
F_ADC_A A

FUNC ADC_A_n
	PCONTB PC,#1
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 1,2

FUNC SUB_HL_
	PCONTB0 HL
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 1,2

	.even
f_sub_a1:
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 1,1	

.macro F_SUB l
FUNC SUB_\l
	GET_\l %d0
	bra f_sub_a1
	.endm

F_SUB B
F_SUB C
F_SUB D
F_SUB E
F_SUB H
F_SUB L
F_SUB A

FUNC SUB_n
	PCONTB PC,#1
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 2,2

FUNC SBC_A_HL_
	PCONTB0 HL
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 1,2

	.even
f_sbc_a1:
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 1,2

.macro F_SBC_A l
FUNC SBC_A_\l
	GET_\l %d0
	bra f_sbc_a1
	.endm

F_SBC_A B
F_SBC_A C
F_SBC_A D
F_SBC_A E
F_SBC_A H
F_SBC_A L
F_SBC_A A

FUNC SBC_A_n
	PCONTB PC,#1
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 1,2

FUNC AND_HL_
	PCONTB0 HL
	GET_A %d7
	exg %d0,%d7
	and.b %d7,%d0
	ANDFLAGS
	SET_A %d0
	RET 1,2

	.even
f_and1:
	GET_A %d0
	and.b %d7,%d0
	ANDFLAGS
	SET_A %d0
	RET 1,2

.macro F_AND l
FUNC AND_\l
	GET_\l %d7
	bra f_and1
	.endm

F_AND B
F_AND C
F_AND D
F_AND E
F_AND H
F_AND L
F_AND A

FUNC AND_n
	PCONTB PC,#1
	GET_A %d7
	exg %d0,%d7
	and.b %d7,%d0
	ANDFLAGS
	SET_A %d0
	RET 2,2

FUNC OR_HL_
	PCONTB0 HL
	GET_A %d7
	exg %d0,%d7
	or.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 1,2

	.even
f_or1:
	GET_A %d0
	or.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 1,2

.macro F_OR l
FUNC OR_\l
	GET_\l %d7
	bra f_or1
	.endm

F_OR B
F_OR C
F_OR D
F_OR E
F_OR H
F_OR L
F_OR A

FUNC OR_n
	PCONTB PC,#1
	GET_A %d7
	exg %d0,%d7
	or.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 2,2

FUNC CP_HL_
	PCONTB0 HL
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	RET 1,2

	.even
f_cp1:
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	RET 1,1

.macro F_CP l
FUNC CP_\l
	GET_\l %d0
	bra f_cp1
	.endm

F_CP B
F_CP C
F_CP D
F_CP E
F_CP H
F_CP L
F_CP A

FUNC CP_n
	PCONTB PC,#1
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	RET 2,2

FUNC JMP_nn
	PCONTW PC,#1
	SET_PC %d0
	RET 0,3

.macro F_JMP f
FUNC JP_nn_\f
	FLAG_\f JMP_nn
	RET 3,3
	.endm

F_JMP NZ
F_JMP Z
F_JMP NC
F_JMP C
F_JMP PO
F_JMP PE
F_JMP P
F_JMP M

FUNC JP_HL_
	S_HL
	S_PC
	move.w %d2,%d4
	E_PC
	E_HL
	RET 0,1

FUNC CALL_nn
	S_PC
	PCONTWR %d4,#1
	move.w %d0,%d7
	E_PC
	S_SP
	sub.w #2,%d4
	move.w %d4,%a0
	E_SP
	S_PC
	add.w #3,%d4
	move.w %d4,%d0
	STOREW
	move.w %d7,%d4
	E_PC
	IRET 4
	RET0

.macro F_CALL f
FUNC CALL_nn_\f
	FLAG_\f CALL_nn
	RET 3,3
	.endm

F_CALL NZ
F_CALL Z
F_CALL NC
F_CALL C
F_CALL PO
F_CALL PE
F_CALL P
F_CALL M

FUNC RET_
	S_SP
	PCONTWR0 %d4
	add.w #2,%d4
	E_SP
	S_PC
	move.b %d0,%d4
	ror.w #8,%d4
	ror.w #8,%d0
	move.b %d0,%d4
	ror.w #8,%d4
	E_PC
	RET 0,3

.macro F_RET f
FUNC RET_\f
	FLAG_\f RET_
	RET 1,1
	.endm

F_RET NZ
F_RET Z
F_RET NC
F_RET C
F_RET PO
F_RET PE
F_RET P
F_RET M

.macro F_RST_p n
FUNC RST_\n
	S_SP
	sub.w #2,%d4
	move.w %d4,%a0
	E_SP
	S_PC
	add.w #1,%d4
	move.w %d4,%d0
	STOREW
	move.w #0x\n,%d4
	E_PC
	RET 0,3 
	.endm

F_RST_p 0
F_RST_p 8
F_RST_p 10
F_RST_p 18
F_RST_p 20
F_RST_p 28
F_RST_p 30
F_RST_p 38

FUNC DI
	clr.b %d0
	SET_IFF1 %d0
	clr.b %d0
	SET_IFF2 %d0
	RET 1,1

FUNC EI
	move.b #1,%d0
	SET_NEI %d0
	RET 1,1

FUNC EX_DE_HL
	GET_DE %d7
	GET_HL %d0
	SET_DE %d0
	SET_HL %d7
	RET 1,1

FUNC EXX
	GET_BC %d0
	GETW REG_BC_,%d7
	SET_BC %d7
	SETW REG_BC_,%d0
	
	GET_DE %d0
	GETW REG_DE_,%d7
	SET_DE %d7
	SETW REG_DE_,%d0
	
	GET_HL %d0
	GETW REG_HL_,%d7
	SET_HL %d7
	SETW REG_HL_,%d0
	RET 1,1

FUNC EX_SP_HL
	PCONTW0 SP
	move.w %d0,%d7
	GET_SP %a0
	GET_HL %d0
	STOREW
	SET_HL %d7
	RET 1,5

FUNC OUT_n_A
	GET_A %d7
	PCONTB PC,#1
	lsl.w #8,%d7
	or.b %d0,%d7
	GET_A %d0
	OUTPORT %d7,%d0
	RET 2,3

FUNC IN_A_n_
	GET_A %d7
	PCONTB PC,#1
	lsl.w #8,%d7
	or.b %d0,%d7
	INPORT %d7
	SET_A %d0
	RET 2,3

FUNC OUTI
	bra noinstr

FUNC IND
	bra noinstr

FUNC OUTD
	bra noinstr

FUNC INIR
	bra noinstr

FUNC OTIR
	bra noinstr

FUNC INDR
	bra noinstr

FUNC OTDR
	bra noinstr

.macro F_LD_HL_r l
FUNC LD_HL_\l
	GET_HL %a0
	GET_\l %d0
	STORE
	RET 1,2
	.endm

F_LD_HL_r B
F_LD_HL_r C
F_LD_HL_r D
F_LD_HL_r E
F_LD_HL_r H
F_LD_HL_r L
F_LD_HL_r A

.macro F_LD_r_HL_ l,e=_HL_
FUNC LD_\l\e
	PCONTB0 HL
	SET_\l %d0
	RET 1,2
	.endm

F_LD_r_HL_ B
F_LD_r_HL_ C
F_LD_r_HL_ D
F_LD_r_HL_ E
F_LD_r_HL_ H
F_LD_r_HL_ L
F_LD_r_HL_ A

.macro F_LD_r_r dl,sl,m=_
FUNC LD_\dl\m\sl
	GET_\sl %d0
	SET_\dl %d0
	RET 1,1
	.endm

F_LD_r_r B,B
F_LD_r_r B,C
F_LD_r_r B,D
F_LD_r_r B,E
F_LD_r_r B,H
F_LD_r_r B,L
F_LD_r_r B,A

F_LD_r_r C,B
F_LD_r_r C,C
F_LD_r_r C,D
F_LD_r_r C,E
F_LD_r_r C,H
F_LD_r_r C,L
F_LD_r_r C,A

F_LD_r_r D,B
F_LD_r_r D,C
F_LD_r_r D,D
F_LD_r_r D,E
F_LD_r_r D,H
F_LD_r_r D,L
F_LD_r_r D,A

F_LD_r_r E,B
F_LD_r_r E,C
F_LD_r_r E,D
F_LD_r_r E,E
F_LD_r_r E,H
F_LD_r_r E,L
F_LD_r_r E,A

F_LD_r_r H,B
F_LD_r_r H,C
F_LD_r_r H,D
F_LD_r_r H,E
F_LD_r_r H,H
F_LD_r_r H,L
F_LD_r_r H,A

F_LD_r_r L,B
F_LD_r_r L,C
F_LD_r_r L,D
F_LD_r_r L,E
F_LD_r_r L,H
F_LD_r_r L,L
F_LD_r_r L,A

F_LD_r_r A,B
F_LD_r_r A,C
F_LD_r_r A,D
F_LD_r_r A,E
F_LD_r_r A,H
F_LD_r_r A,L
F_LD_r_r A,A

FUNC HALT_
	GETL TIME_INT,%d0
	move.l #20000,%d7
	sub.l %d0,%d7
	SET_time %d7
	RET 1,0

FUNC LD_HL_n
	PCONTB PC,#1
	GET_HL %a0
	STORE
	RET 2,2

.macro F_LD_r_n l,m=_n
FUNC LD_\l\m
	PCONTB PC,#1
	SET_\l %d0
	RET 2,2
	.endm

F_LD_r_n B
F_LD_r_n C
F_LD_r_n D
F_LD_r_n E
F_LD_r_n H
F_LD_r_n L
F_LD_r_n A

.macro F_LD_dd_nn n,m=_nn
FUNC LD_\n\m
	PCONTW PC,#1
	SET_\n %d0
	RET 3,3
	.endm

F_LD_dd_nn BC
F_LD_dd_nn DE
F_LD_dd_nn HL
F_LD_dd_nn SP

FUNC LD_HL_nn_o
	PCONTW PC,#1
	PCONTWR0 %d0
	SET_HL %d0
	RET 3,4


FUNC JR_e
	PCONTB PC,#1
	S_PC
	move.w %d4,%a0
	add.w #2,%a0
	ext.w %d0
	add.w %d0,%a0
	move.w %a0,%d4
	E_PC
	RET 0,3

.macro F_JR_c_e f,m=_e
FUNC JR_\f\m
	FLAG_\f JR_e
	RET 2,2
	.endm

F_JR_c_e Z
F_JR_c_e NZ
F_JR_c_e C
F_JR_c_e NC

FUNC DJNZ_e
	S_B
	sub.b #1,%d1
	move.b %d1,%d0
	E_B
	cmp.b #0,%d0
	bne JR_e
	RET 2,3

FUNC LD_A_BC_
	PCONTB0 BC
	SET_A %d0
	RET 1,2

FUNC LD_A_DE_
	PCONTB0 DE
	SET_A %d0
	RET 1,2

FUNC LD_A_nn_
	PCONTW PC,#1
	move.w %d0,%a0
	READB0
	SET_A %d0
	RET 3,4

FUNC LD_BC_A
	GET_BC %a0
	GET_A %d0
	STORE
	RET 1,2

FUNC LD_DE_A
	GET_DE %a0
	GET_A %d0
	STORE
	RET 1,2

FUNC LD_nn_A
	PCONTW PC,#1
	move.w %d0,%a0
	GET_A %d0
	STORE
	RET 3,3

FUNC LD_nn_HL2
	PCONTW PC,#1
	move.w %d0,%a0
	GET_HL %d0
	STOREW
	RET 3,5

FUNC INC_HL_
	PCONTB0 HL
	move.b #1,%d7
	ADDFLAGS
	GET_HL %a0
	STORE
	RET 1,3

.macro F_INC_r l
FUNC INC_\l
	GET_\l %d0
	move.b #1,%d7
	ADDFLAGS
	SET_\l %d0
	RET 1,1
	.endm

F_INC_r B
F_INC_r C
F_INC_r D
F_INC_r E
F_INC_r H
F_INC_r L
F_INC_r A

FUNC DEC_HL_
	PCONTB0 HL
	move.b #1,%d7
	SUBFLAGS
	GET_HL %a0
	STORE
	RET 1,3

.macro F_DEC_r l
FUNC DEC_\l
	GET_\l %d0
	move.b #1,%d7
	SUBFLAGS
	SET_\l %d0
	RET 1,1
	.endm

F_DEC_r B
F_DEC_r C
F_DEC_r D
F_DEC_r E
F_DEC_r H
F_DEC_r L
F_DEC_r A

.macro F_INC_ss hl
FUNC INC_\hl
	ADD_\hl #1
	RET 1,2
	.endm

F_INC_ss BC
F_INC_ss DE
F_INC_ss HL
F_INC_ss SP

.macro F_DEC_ss hl
FUNC DEC_\hl
	SUB_\hl #1
	RET 1,2
	.endm

F_DEC_ss BC
F_DEC_ss DE
F_DEC_ss HL
F_DEC_ss SP

	.even
f_add_hl_ss1:
	GET_HL %d7
	ADDHFLAGS
	SET_HL %d0
	RET 1,2

.macro F_ADD_HL_ss hl
FUNC ADD_HL_\hl
	GET_\hl %d0
	bra f_add_hl_ss1
	.endm

F_ADD_HL_ss BC
F_ADD_HL_ss DE
F_ADD_HL_ss HL
F_ADD_HL_ss SP

FUNC DAA_
	GET_A %d0
	move.w %d0,%a0
	FLAG_H daa5
	and.b #0x0f,%d0
	cmp.b #0x0a,%d0
	bcs daa1
daa5:
	move.w %a0,%d0
	add.b #0x06,%d0
	STFLAGSOC
	move.w %d0,%a0
	FLAG1_H
	bra daa3
	.even
daa1:
	FLAG0_H
daa3:
	move.w %a0,%d0
	FLAG_C daa6
	cmp.b #0xa0,%d0
	bcs daa2
daa6:
	add.b #0x60,%d0
	move.w %d0,%a0
	FLAG1_C
	bra daa4
	.even
daa2:
	FLAG0_C
daa4:
	move.w %a0,%d0
	cmp.b #0,%d0
	STFLAGSC %d7
	TESTFN P,%d0,0
	SET_A %d0
	RET 1,1

FUNC CPL
	S_A
	not.b %d1
	E_A 
	FLAG1_H
	FLAG1_N
	RET 1,1

FUNC CCF
	FLAG_C ccf1
	FLAG0_H
	bra ccf2
	.even
ccf1:
	FLAG1_H
ccf2:
	FLAG0_N
	FLAGC_C
	RET 1,1

FUNC SCF
	FLAG0_H
	FLAG0_N
	FLAG1_C
	RET 1,1

FUNC RLCA
	GET_A %d0
	rol.b #1,%d0
	STFLAGSOC
rlca1:
	FLAG0_H
	FLAG0_N
	SET_A %d0
	RET 1,1

FUNC RLA
	GET_A %d0
	SETCX %d7
	roxl.b #1,%d0
	STFLAGSOC
	bra rlca1

FUNC RRCA
	GET_A %d0
	ror.b #1,%d0
	STFLAGSOC
	bra rlca1

FUNC RRA
	GET_A %d0
	SETCX %d7
	roxr.b #1,%d0
	STFLAGSOC
	bra rlca1

FUNC EX_AF_AF_
	FLAGS_TO_F
	GET_AF %d0
	move.w %d0,%d7
	GETW REG_AF_,%d0
	SET_AF %d0
	SETW REG_AF_,%d7
	F_TO_FLAGS
	RET 1,1

FUNC LD_A_I
	GET_I %d0
	cmp.b #0,%d0
	STFLAGSC %d7
ldai1:
	FLAG0_H
	FLAG0_N
	GET_IFF2 %d7
	cmp.b #0,%d7
	beq ldai2
	FLAG1_P
	bra ldai3
	.even
ldai2:
	FLAG0_P
ldai3:
	SET_A %d0
	RET 2,2

FUNC LD_A_R
	GET_R %d0
	cmp.b #0,%d0
	STFLAGSC %d7
	bra ldai1

FUNC LD_I_A
	GET_A %d0
	SET_I %d0
	RET 2,2

FUNC LD_R_A
	GET_A %d0
	SET_R %d0
	RET 2,2

.macro F_LD_dd_nn_ n,m=_nn_
FUNC LD_\n\m
	PCONTW PC,#2
	PCONTWR0 %d0
	SET_\n %d0
	RET 4,5
	.endm

F_LD_dd_nn_ BC
F_LD_dd_nn_ DE
F_LD_dd_nn_ HL
F_LD_dd_nn_ SP

.macro F_LD_nn_dd n
FUNC LD_nn_\n
	PCONTW PC,#2
	move.w %d0,%a0
	GET_\n %d0
	STOREW
	RET 4,5
	.endm

F_LD_nn_dd BC
F_LD_nn_dd DE
F_LD_nn_dd HL
F_LD_nn_dd SP

	.even
f_adc_hl_ss1:
	GET_FLAGS %d7
	and.w #0x01,%d7
	add.w %d7,%d0
	GET_HL %d7
	ADDHFLAGSC
	SET_HL %d0
	RET 2,4

.macro F_ADC_HL_ss hl
FUNC ADC_HL_\hl
	GET_\hl %d0
	bra f_adc_hl_ss1
	.endm

F_ADC_HL_ss BC
F_ADC_HL_ss DE
F_ADC_HL_ss HL
F_ADC_HL_ss SP

	.even
f_sbc_hl_ss1:
	GET_FLAGS %d7
	and.w #0x01,%d7
	add.w %d7,%d0
	GET_HL %d7
	exg %d7,%d0
	SUBHFLAGS
	SET_HL %d0
	RET 2,4

.macro F_SBC_HL_ss hl
FUNC SBC_HL_\hl
	GET_\hl %d0
	bra f_sbc_hl_ss1
	.endm

F_SBC_HL_ss BC
F_SBC_HL_ss DE
F_SBC_HL_ss HL
F_SBC_HL_ss SP

FUNC RETI
	PCONTW0 SP
	SET_PC %d0
	ADD_SP #2
	RET 2,4

FUNC RETN
	PCONTW0 SP
	SET_PC %d0
	ADD_SP #2
	GET_IFF2 %d0
	SET_IFF1 %d0
	RET 2,4

FUNC NEG_
	GET_A %d0
	cmp.b #0x80,%d0
	beq neg_1
	FLAG0_P
	bra neg_2
	.even
neg_1:
	FLAG1_P
neg_2:
	cmp.b #0,%d0
	bne neg_3
	FLAG1_C
	bra neg_4
	.even
neg_3:
	FLAG0_C
neg_4:
	neg.b %d0
	STFLAGSCP %d7
	FLAG1_H
	FLAG1_N
	SET_A %d0
	RET 2,2

FUNC IM_0
	clr.b %d0
	SET_IM %d0
	RET 2,2

FUNC IM_1
	move.b #1,%d0
	SET_IM %d0
	RET 2,2

FUNC IM_2
	move.b #2,%d0
	SET_IM %d0
	RET 2,2

FUNC RLD
	move.l %d5,-(%sp)
	move.l %d6,-(%sp)
	# %d5 = a
	GET_A %d5
	and.b #0x0f,%d5
	PCONTB0 HL
	# %d0 = hl
	move.b %d0,%d6
	and.b #0xf0,%d6
	# %d6 = h
	move.b %d0,%d7
	and.b #0x0f,%d7
	# %d7 = l
	lsl.b #4,%d7
	move.b %d5,%d0
	or.b %d7,%d0
	GET_HL %d5
	move.w %d5,%a0
	STORE
	lsr.b #4,%d6
	GET_A %d0
	and.b #0xf0,%d0
	or.b %d6,%d0
	SET_A %d0
	cmp.b #0,%d0
	STFLAGSC %d7
	FLAG0_N
	TESTFN P,%d0,0
	FLAG0_H
	move.l (%sp)+,%d6
	move.l (%sp)+,%d5
	RET 2,5

FUNC RRD
	move.l %d5,-(%sp)
	move.l %d6,-(%sp)
	# %d5 = a
	GET_A %d5
	and.b #0x0f,%d5
	PCONTB0 HL
	# %d0 = hl
	move.b %d0,%d6
	and.b #0xf0,%d6
	# %d6 = h
	move.b %d0,%d7
	and.b #0x0f,%d7
	# %d7 = l
	lsl.b #4,%d5
	lsr.b #4,%d6
	move.b %d5,%d0
	or.b %d6,%d0
	GET_HL %d5
	move.w %d5,%a0
	STORE
	GET_A %d0
	and.b #0xf0,%d0
	or.b %d7,%d0
	SET_A %d0
	cmp.b #0,%d0
	STFLAGSC %d7
	FLAG0_N
	TESTFN P,%d0,0
	FLAG0_H
	move.l (%sp)+,%d6
	move.l (%sp)+,%d5
	RET 2,5

FUNC LDI
	PCONTB0 HL
	GET_DE %a0
	STORE
	ADD_DE #1
	ADD_HL #1
	SUB_BC #1
	FLAG0_H
	FLAG0_N
	GET_BC %d0
	cmp.w #0,%d0
	bne ldi1
	FLAG0_P
	bra ldi2
ldi1:
	FLAG1_P
ldi2:
	RET 2,4

FUNC LDIR
	SRET 2
ldir0:
	PCONTB0 HL
	GET_DE %a0
	STORE
	ADD_DE #1
	ADD_HL #1
	SUB_BC #1
	GET_BC %d0
	cmp.w #0,%d0
	bne ldir1
	FLAG0_H
	FLAG0_P
	FLAG0_N
	IRET 4
	bra ldir2
ldir1:
	IRET 5
	bra ldir0
ldir2:
	RET 0,0

FUNC LDD
	PCONTB0 HL
	GET_DE %a0
	STORE
	SUB_DE #1
	SUB_HL #1
	SUB_BC #1
	FLAG0_H
	FLAG0_N
	GET_BC %d0
	cmp.w #0,%d0
	bne ldd1
	FLAG0_P
	bra ldd2
ldd1:
	FLAG1_P
ldd2:
	RET 2,4

FUNC LDDR
	SRET 2
lddr0:
	PCONTB0 HL
	GET_DE %a0
	STORE
	SUB_DE #1
	SUB_HL #1
	SUB_BC #1
	GET_BC %d0
	cmp.w #0,%d0
	bne lddr1
	FLAG0_H
	FLAG0_P
	FLAG0_N
	IRET 4
	bra lddr2
lddr1:
	IRET 5
	bra lddr0
lddr2:
	RET 0,0

FUNC CPI
	ADD_HL #1
	SUB_BC #1
	PCONTB0 HL
	move.b %d0,%d7
	GET_A %d0
	SUBFLAGS
	FLAG1_N
	GET_BC %d0
	cmp.w #0,%d0
	bne cpi1
	FLAG0_P
	bra cpi2
cpi1:
	FLAG1_P
cpi2:
	RET 2,4

FUNC CPIR
	SRET 2
cpir0:
	ADD_HL #1
	SUB_BC #1
	PCONTB0 HL
	move.b %d0,%d7
	GET_A %d0
	cmp.b %d7,%d0
	beq cpir1
	GET_BC %a0
	cmp.w #0,%a0
	beq cpir1
	IRET 5
	bra cpir0
cpir1:
	SUBFLAGS
	FLAG1_N
	GET_BC %d0
	cmp.w #0,%d0
	bne cpir2
	FLAG0_P
	bra cpir3
cpir2:
	FLAG1_P
cpir3:
	IRET 2
	RET 0,0

FUNC CPD
	SUB_HL #1
	SUB_BC #1
	PCONTB0 HL
	move.b %d0,%d7
	GET_A %d0
	SUBFLAGS
	FLAG1_N
	GET_BC %d0
	cmp.w #0,%d0
	bne cpd1
	FLAG0_P
	bra cpd2
cpd1:
	FLAG1_P
cpd2:
	RET 2,4

FUNC CPDR
	SRET 2
cpdr0:
	SUB_HL #1
	SUB_BC #1
	PCONTB0 HL
	move.b %d0,%d7
	GET_A %d0
	cmp.b %d7,%d0
	beq cpdr1
	GET_BC %a0
	cmp.w #0,%a0
	beq cpdr1
	IRET 5
	bra cpdr0
cpdr1:
	SUBFLAGS
	FLAG1_N
	GET_BC %d0
	cmp.w #0,%d0
	bne cpdr2
	FLAG0_P
	bra cpdr3
cpdr2:
	FLAG1_P
cpdr3:
	IRET 2
	RET 0,0

	.even
f_in_r_c1:
	cmp.b #0,%d0
	STFLAGSC %d7
	TESTFN P,%d0,0
	FLAG0_H
	FLAG0_N
	RET 2,3

.macro F_IN_r_C l,m=_C_
FUNC IN_\l\m
	GET_BC %a0
	INPORT %a0
	SET_\l %d0
	bra f_in_r_c1
	.endm

F_IN_r_C B
F_IN_r_C C
F_IN_r_C D
F_IN_r_C E
F_IN_r_C H
F_IN_r_C L
F_IN_r_C A

.macro F_OUT_C_r l
FUNC OUT_C_\l
	GET_BC %a0
	GET_\l %d0
	OUTPORT %a0,%d0
	RET 2,3
	.endm

F_OUT_C_r B
F_OUT_C_r C
F_OUT_C_r D
F_OUT_C_r E
F_OUT_C_r H
F_OUT_C_r L
F_OUT_C_r A

FUNC INI 
	GET_BC %a0
	INPORT %a0
	GET_HL %d7
	move.w %d7,%a0
	STORE
	ADD_HL #1
	SUB_B #1
	GET_B %d0
	cmp.b #0,%d0
	bne ini1
	FLAG1_Z
	bra ini2
	.even
ini1:
	FLAG0_Z
ini2:
	RET 2,4

bit0:
	FLAG1_Z
	FLAG1_P
	FLAG0_S
	bra bite
	.even
bit1:
	FLAG0_Z
	FLAG0_P
	cmp.w #7,%a0
	bne bit3
	FLAG1_S
	bra bite
	.even
bit3:
	FLAG0_S
bite:
	FLAG1_H
	FLAG0_N
	RET 2,5

.macro F_BIT_b_HL_ n,m=_HL_
FUNC BIT_\n\m
	PCONTB0 HL
	move.w #\n,%a0
	btst.b #\n,%d0
	beq bit0
	bra bit1
	.endm

F_BIT_b_HL_ 0
F_BIT_b_HL_ 1
F_BIT_b_HL_ 2
F_BIT_b_HL_ 3
F_BIT_b_HL_ 4
F_BIT_b_HL_ 5
F_BIT_b_HL_ 6
F_BIT_b_HL_ 7

.macro F_BIT_b_r n,l,m=_
FUNC BIT_\n\m\l
	move.w #\n,%a0
	GET_\l %d0
	btst.b #\n,%d0
	beq bit0
	bra bit1
	.endm

F_BIT_b_r 0,B
F_BIT_b_r 1,B
F_BIT_b_r 2,B
F_BIT_b_r 3,B
F_BIT_b_r 4,B
F_BIT_b_r 5,B
F_BIT_b_r 6,B
F_BIT_b_r 7,B

F_BIT_b_r 0,C
F_BIT_b_r 1,C
F_BIT_b_r 2,C
F_BIT_b_r 3,C
F_BIT_b_r 4,C
F_BIT_b_r 5,C
F_BIT_b_r 6,C
F_BIT_b_r 7,C

F_BIT_b_r 0,D
F_BIT_b_r 1,D
F_BIT_b_r 2,D
F_BIT_b_r 3,D
F_BIT_b_r 4,D
F_BIT_b_r 5,D
F_BIT_b_r 6,D
F_BIT_b_r 7,D

F_BIT_b_r 0,E
F_BIT_b_r 1,E
F_BIT_b_r 2,E
F_BIT_b_r 3,E
F_BIT_b_r 4,E
F_BIT_b_r 5,E
F_BIT_b_r 6,E
F_BIT_b_r 7,E

F_BIT_b_r 0,H
F_BIT_b_r 1,H
F_BIT_b_r 2,H
F_BIT_b_r 3,H
F_BIT_b_r 4,H
F_BIT_b_r 5,H
F_BIT_b_r 6,H
F_BIT_b_r 7,H

F_BIT_b_r 0,L
F_BIT_b_r 1,L
F_BIT_b_r 2,L
F_BIT_b_r 3,L
F_BIT_b_r 4,L
F_BIT_b_r 5,L
F_BIT_b_r 6,L
F_BIT_b_r 7,L

F_BIT_b_r 0,A
F_BIT_b_r 1,A
F_BIT_b_r 2,A
F_BIT_b_r 3,A
F_BIT_b_r 4,A
F_BIT_b_r 5,A
F_BIT_b_r 6,A
F_BIT_b_r 7,A

	.even
f_set_b_hl_1:
	GET_HL %a0
	STORE 
	RET 2,4

.macro F_SET_b_HL_ n,m=_HL_
FUNC SET_\n\m
	PCONTB0 HL
	bset.b #\n,%d0
	bra f_set_b_hl_1
	.endm

F_SET_b_HL_ 0
F_SET_b_HL_ 1
F_SET_b_HL_ 2
F_SET_b_HL_ 3
F_SET_b_HL_ 4
F_SET_b_HL_ 5
F_SET_b_HL_ 6
F_SET_b_HL_ 7

.macro F_SET_b_r n,l,r,m=_
FUNC SET_\n\m\l
	S_\l
	bset.b #\n,\r
	E_\l
	RET 2,2
	.endm

F_SET_b_r 0,B,%d1
F_SET_b_r 1,B,%d1
F_SET_b_r 2,B,%d1
F_SET_b_r 3,B,%d1
F_SET_b_r 4,B,%d1
F_SET_b_r 5,B,%d1
F_SET_b_r 6,B,%d1
F_SET_b_r 7,B,%d1

F_SET_b_r 0,C,%d1
F_SET_b_r 1,C,%d1
F_SET_b_r 2,C,%d1
F_SET_b_r 3,C,%d1
F_SET_b_r 4,C,%d1
F_SET_b_r 5,C,%d1
F_SET_b_r 6,C,%d1
F_SET_b_r 7,C,%d1

F_SET_b_r 0,D,%d2
F_SET_b_r 1,D,%d2
F_SET_b_r 2,D,%d2
F_SET_b_r 3,D,%d2
F_SET_b_r 4,D,%d2
F_SET_b_r 5,D,%d2
F_SET_b_r 6,D,%d2
F_SET_b_r 7,D,%d2

F_SET_b_r 0,E,%d2
F_SET_b_r 1,E,%d2
F_SET_b_r 2,E,%d2
F_SET_b_r 3,E,%d2
F_SET_b_r 4,E,%d2
F_SET_b_r 5,E,%d2
F_SET_b_r 6,E,%d2
F_SET_b_r 7,E,%d2

F_SET_b_r 0,H,%d2
F_SET_b_r 1,H,%d2
F_SET_b_r 2,H,%d2
F_SET_b_r 3,H,%d2
F_SET_b_r 4,H,%d2
F_SET_b_r 5,H,%d2
F_SET_b_r 6,H,%d2
F_SET_b_r 7,H,%d2

F_SET_b_r 0,L,%d2
F_SET_b_r 1,L,%d2
F_SET_b_r 2,L,%d2
F_SET_b_r 3,L,%d2
F_SET_b_r 4,L,%d2
F_SET_b_r 5,L,%d2
F_SET_b_r 6,L,%d2
F_SET_b_r 7,L,%d2

F_SET_b_r 0,A,%d1
F_SET_b_r 1,A,%d1
F_SET_b_r 2,A,%d1
F_SET_b_r 3,A,%d1
F_SET_b_r 4,A,%d1
F_SET_b_r 5,A,%d1
F_SET_b_r 6,A,%d1
F_SET_b_r 7,A,%d1

.macro F_RES_b_HL_ n,m=_HL_
FUNC RES_\n\m
	PCONTB0 HL
	bclr.b #\n,%d0
	bra f_set_b_hl_1
	.endm

F_RES_b_HL_ 0
F_RES_b_HL_ 1
F_RES_b_HL_ 2
F_RES_b_HL_ 3
F_RES_b_HL_ 4
F_RES_b_HL_ 5
F_RES_b_HL_ 6
F_RES_b_HL_ 7

.macro F_RES_b_r n,l,r,m=_
FUNC RES_\n\m\l
	S_\l
	bclr.b #\n,\r
	E_\l
	RET 2,2
	.endm

F_RES_b_r 0,B,%d1
F_RES_b_r 1,B,%d1
F_RES_b_r 2,B,%d1
F_RES_b_r 3,B,%d1
F_RES_b_r 4,B,%d1
F_RES_b_r 5,B,%d1
F_RES_b_r 6,B,%d1
F_RES_b_r 7,B,%d1

# --------------------------------------------------------------------------------------

	.even
readb:
	# %a0 = z80dir
	# %d0 = desp
	# %a2 = ret addr
	# exit: %d0 = code
	ext.w %d0
	add.w %d0,%a0
	# %a0 = z80dir
	# %a2 = ret addr
	# exit: %d0 = code
readb0:
	move.w %a0,%d0
	and.w #0xc000,%d0
	bne readb01
	move.w %a0,%d0
	and.l #0xffff,%d0
	move.b (%a3,%d0.l),%d0
	jmp (%a2)
	.even
readb01:
	move.w %a0,%d0
	sub.w #0x4000,%d0
	and.l #0xffff,%d0
	move.b (%a4,%d0.l),%d0
	jmp (%a2)

	.even
readw:
	# %a0 = z80dir
	# %d0 = desp
	# %a2 = ret addr
	# exit: %d0 = code
	ext.w %d0
	add.w %d0,%a0
	# %a0 = z80dir
	# %a2 = ret addr
	# exit: %d0 = code
readw0:
	move.w %a0,%d0
	and.w #0xc000,%d0
	bne readw01
	move.w %a0,%d0
	and.l #0xffff,%d0
	cmp.w #0x3fff,%d0
	beq readw02
	btst.b #0,%d0
	beq readw04
	sub.w #1,%d0
	move.l (%a3,%d0.l),%d0
	lsr.l #8,%d0
	ror.w #8,%d0
	jmp (%a2)
	.even
readw04:
	move.w (%a3,%d0.l),%d0
	ror.w #8,%d0
	jmp (%a2)
	.even
readw02:
	move.b (%a4),%d0
	lsl.w #8,%d0
	move.l #0x3fff,%d7
	move.b (%a3,%d7.l),%d0
	jmp (%a2)
	.even
readw01:
	move.w %a0,%d0
	sub.w #0x4000,%d0
	and.l #0xffff,%d0
	cmp.w #0xbfff,%d0
	beq readw03
	btst.b #0,%d0
	beq readw05
	sub.w #1,%d0
	move.l (%a4,%d0.l),%d0
	lsr.l #8,%d0
	ror.w #8,%d0
	jmp (%a2)
	.even
readw05:
	move.w (%a4,%d0.l),%d0
	ror.w #8,%d0
	jmp (%a2)
	.even
readw03:
	move.b (%a3),%d0
	lsl.w #8,%d0
	move.l #0xbfff,%d7
	move.b (%a4,%d7.l),%d0
	jmp (%a2)

	# debug
	.even
t_mem:
	.ascii "%04Lx = %02Lx\0"
	# debug

	.even
store0:
	# %a0 = dir
	# %d0 = code
	# %a2 = return addr
	move.w %d0,%a1
	move.w %a0,%d0
	and.w #0xc000,%d0
	beq f_store0
	move.w %a0,%d0
	sub.w #0x4000,%d0
	and.l #0xffff,%d0
	exg %d0,%a1
	move.b %d0,(%a4,%a1.l)	
	jmp (%a2)
	.even
f_store0:
	move.w %a1,%d0
	jmp (%a2)

.macro SCR_BIT b
	#void putpixel (USHORT x,USHORT y,UCHAR bc,UCHAR bb,UCHAR code,UCHAR mask)
	#PUTPIXEL %d2,%d7,%d5,%d4,%d0,%d1
	btst.b #\b,%d1
	beq scrbite\@
	GETL sc,%a0
	move.w %d2,%d6
	lsr.w #1,%d6
	and.l #0xffff,%d6
	add.l %d6,%a0
	move.w %d7,%d6
	mulu.w #80,%d6
	and.l #0xffff,%d6
	add.l %d6,%a0
	move.b (%a0),%d6
	move.w %d7,%a1
	btst.b #\b,%d0
	bne scrbit0\@
	move.b %d4,%d7
	bra scrbit2\@
scrbit0\@:
	move.b %d5,%d7
scrbit2\@:
	btst.b #0,%d2
	beq scrbit1\@
	and.b #0xf0,%d6
	or.b %d7,%d6
	bra scrbito\@
scrbit1\@:
	and.b #0x0f,%d6
	lsl.b #4,%d7
	or.b %d7,%d6
scrbito\@:
	move.b %d6,(%a0)
	move.w %a1,%d7
scrbite\@:
	add.w #1,%d2
	.endm

	.even
screenmem:
	# %a0 = z80 dir
	# %d0 = code
	# %d1 = mask
	# %a1 saved
	#
	# 00000000 00000000
	#    ||| | | ||   |
	#    m  f   s   x
	# x = 0-31 x pos			and 0x001f
	# s = 0-7  subscreen/8		and 0x00e0	
	# f = 0-7  y subscreen/8	and 0x0700
	# m = 0-2  subscreen		and 0x1800
	#
	# 00000000 00000000
	#       || | ||   |
	#       m   s   x
	# x = 0-31 x pos			and 0x001f
	# s = 0-7  subscreen/8		and 0x00e0
	# m = 0-2  subscreen        and 0x0300
	#
	/*
	move.l %d2,-(%sp)
	move.l %d7,-(%sp)
	move.l %d4,-(%sp)
	move.l %d5,-(%sp)
	move.l %d6,-(%sp)
	move.l %a0,-(%sp)
	*/
	PUSHALL
	
	# x = %d2;
	move.w %a0,%d2
	and.w #0x001f,%d2
	# s = %d7;
	move.w %a0,%d7
	and.w #0x00e0,%d7
	# f = %d4;
	move.w %a0,%d4
	and.w #0x0700,%d4
	# m = %d5;
	move.w %a0,%d5
	and.w #0x1800,%d5
	# col = %a0
	clr.l %d6
	or.w %d2,%d6
	or.w %d7,%d6
	lsr.w #3,%d5
	or.w %d5,%d6
	add.w #0x1800,%d6
	move.l %d6,%a0
	# x = x << 3
	lsl.w #3,%d2
	# y = (m<<6) + (s<<3) + f;
	lsr.w #2,%d5
	lsr.w #2,%d7
	lsr.w #8,%d4
	add.w %d5,%d7
	add.w %d4,%d7
	# x = %d2
	# y = %d7
	# bn = GMEM(s_screencol[mem-0x4000]);
	move.b (%a4,%a0.l),%d6
	# bb = (bn&0x38) >> 3;
	move.b %d6,%d4
	and.b #0x38,%d4
	lsr.b #3,%d4
	# bc = (bn&0x07);
	move.b %d6,%d5
	and.b #0x07,%d5
	# blink
	move.w %d6,%a0
	and.b #0x40,%d6
	beq scrnob
	add.b #8,%d4
	add.b #8,%d5
scrnob:
	move.w %a0,%d6
	# flash
	and.b #0x80,%d6
	beq scrnof
	GET_flash %d6
	cmp.b #0,%d6
	beq scrnof
	exg %d4,%d5
scrnof:
	/*
	and.w #0x00ff,%d2
	and.w #0x00ff,%d7
	GETW vx,%d6
	sub.w %d6,%d2
	GETW vy,%d6
	sub.w %d6,%d7
	# if (x>=0 && x<160 && y>=0 && y<160)
	cmp.w #0,%d2
	bmi noscr
	cmp.w #160,%d2
	bcc noscr
	beq noscr
	cmp.w #0,%d7
	bmi noscr
	cmp.b #160,%d7
	bcc noscr
	beq noscr
	SCR_BIT 7
	SCR_BIT 6
	SCR_BIT 5
	SCR_BIT 4
	SCR_BIT 3
	SCR_BIT 2
	SCR_BIT 1
	SCR_BIT 0
noscr:
	*/
	/* PUSHALL */
	# PUTPIXEL x,y,bc,bb,code,mask
	
	PUTPIXEL %d2,%d7,%d5,%d4,%d0,%d1 
	
	/* POPALL */
	
	POPALL
	/*
	move.l (%sp)+,%a0
	move.l (%sp)+,%d6
	move.l (%sp)+,%d5
	move.l (%sp)+,%d4
	move.l (%sp)+,%d7
	move.l (%sp)+,%d2
	*/
	rts


	.even
colormem:
	# %a0 = z80 dir
	# %a1 saved
	#
	# 00000000 00000000
	#    ||| | | ||   |
	#    m  f   s   x
	# x = 0-31 x pos			and 0x001f
	# s = 0-7  subscreen/8		and 0x00e0	
	# f = 0-7  y subscreen/8	and 0x0700
	# m = 0-2  subscreen		and 0x1800
	#
	# 00000000 00000000
	#       || | ||   |
	#       m   s   x
	# x = 0-31 x pos			and 0x001f
	# s = 0-7  subscreen/8		and 0x00e0
	# m = 0-2  subscreen        and 0x0300
	#
	move.l %d0,-(%sp)
	move.l %d1,-(%sp)
	move.l %d2,-(%sp)
	move.l %a2,-(%sp)
	move.l %a0,-(%sp)
	# x | s = %d0
	move.w %a0,%d0
	and.w #0x00ff,%d0
	# m = %d1
	move.w %a0,%d1
	and.w #0x0300,%d1
	lsl.w #3,%d1
	or.w %d1,%d0
	and.l #0xffff,%d0
	move.l %a4,%a2
	add.l %d0,%a2
	add.w #0x4000,%d0
	move.w %d0,%a0
	clr.b %d2
colmc:
	move.b (%a2),%d0
	move.b #0xff,%d1
	lea screenmem(%pc),%a1
	jsr (%a1)
	add.w #256,%a0
	add.l #256,%a2
	add.b #1,%d2
	cmp.b #8,%d2
	bcs colmc
	move.l (%sp)+,%a0
	move.l (%sp)+,%a2
	move.l (%sp)+,%d2
	move.l (%sp)+,%d1
	move.l (%sp)+,%d0
	rts


	.even
addflagsc:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	ADDFLAG_H
	add.b %d7,%d0
	STFLAGS %d7
	FLAG0_N
	jmp (%a2)

	.even
addflags:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	ADDFLAG_H
	add.b %d7,%d0
	STFLAGSC %d7
	FLAG0_N
	jmp (%a2)

	.even
addhflagsc:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	ADDHFLAG_H
	add.w %d7,%d0
	STFLAGS %d7
	FLAG0_N
	jmp (%a2)

	.even
addhflags:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	ADDHFLAG_H
	add.w %d7,%d0
	STFLAGSOC
	FLAG0_N
	jmp (%a2)

	.even
subflagsc:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	SUBFLAG_H
	sub.b %d7,%d0
	STFLAGS %d7
	FLAG1_N
	jmp (%a2)

	.even
subflags:
	# %d0 = op1
	# %d7 = op2
	# %d0 = result
	SUBFLAG_H
	sub.b %d7,%d0
	STFLAGSC %d7
	FLAG1_N
	jmp (%a2)

	.even
subhflags:
	SUBHFLAG_H
	sub.w %d7,%d0
	STFLAGS %d7
	FLAG1_N
	jmp (%a2)

.macro TESTPORT p,num
	move.w %d0,%d2
	and.w #\p,%d2
	bne testp\@
	GETB KEYBOARD\num,%d2
	and.b %d2,%d1
testp\@:
	.endm

inport:
	# %d0 = port
	move.l %d1,-(%sp)
	move.l %d2,-(%sp)
	move.b #0xff,%d1
	cmp.b #0xfe,%d0
	bne kport
	TESTPORT 0x0100,0
	TESTPORT 0x0200,1
	TESTPORT 0x0400,2
	TESTPORT 0x0800,3
	TESTPORT 0x1000,4
	TESTPORT 0x2000,5
	TESTPORT 0x4000,6
	TESTPORT 0x8000,7
	bra noport
	.even
kport:
	cmp.b #0x1f,%d0
	bne noport
	GETB KEMPSTON,%d1
noport:
	move.b %d1,%d0
	move.l (%sp)+,%d2
	move.l (%sp)+,%d1
	jmp (%a2)


	.even
	.globl	Init
Init:
	link.w %a6,#0
	PUSHALL
	LOAD_OPCODE opcode,noinstr
	LOAD_OPCODE opcodeed,NOPNOP
	LOAD_OPCODE opcodecb,noinstr
	LOAD_OPCODE opcodedd,NOP
	LOAD_OPCODE opcodefd,NOP
	LOAD_OPCODE opcodeddcb,noinstr
	LOAD_OPCODE opcodefdcb,noinstr
	# opcode
	lea opcode@END.w(%a5),%a0
	SET_OPCODE 0x00,NOP	
	SET_OPCODE 0x01,LD_BC_nn
	SET_OPCODE 0x02,LD_BC_A
	SET_OPCODE 0x03,INC_BC
	SET_OPCODE 0x04,INC_B
	SET_OPCODE 0x05,DEC_B
	SET_OPCODE 0x06,LD_B_n
	SET_OPCODE 0x07,RLCA
	SET_OPCODE 0x08,EX_AF_AF_
	SET_OPCODE 0x09,ADD_HL_BC
	SET_OPCODE 0x0a,LD_A_BC_
	SET_OPCODE 0x0b,DEC_BC
	SET_OPCODE 0x0c,INC_C
	SET_OPCODE 0x0d,DEC_C
	SET_OPCODE 0x0e,LD_C_n
	SET_OPCODE 0x0f,RRCA
	SET_OPCODE 0x10,DJNZ_e
	SET_OPCODE 0x11,LD_DE_nn
	SET_OPCODE 0x12,LD_DE_A
	SET_OPCODE 0x13,INC_DE
	SET_OPCODE 0x14,INC_D
	SET_OPCODE 0x15,DEC_D
	SET_OPCODE 0x16,LD_D_n
	SET_OPCODE 0x17,RLA
	SET_OPCODE 0x18,JR_e
	SET_OPCODE 0x19,ADD_HL_DE
	SET_OPCODE 0x1a,LD_A_DE_
	SET_OPCODE 0x1b,DEC_DE
	SET_OPCODE 0x1c,INC_E
	SET_OPCODE 0x1d,DEC_E
	SET_OPCODE 0x1e,LD_E_n
	SET_OPCODE 0x1f,RRA
	SET_OPCODE 0x20,JR_NZ_e
	SET_OPCODE 0x21,LD_HL_nn
	SET_OPCODE 0x22,LD_nn_HL2
	SET_OPCODE 0x23,INC_HL
	SET_OPCODE 0x24,INC_H
	SET_OPCODE 0x25,DEC_H
	SET_OPCODE 0x26,LD_H_n
	SET_OPCODE 0x27,DAA_
	SET_OPCODE 0x28,JR_Z_e
	SET_OPCODE 0x29,ADD_HL_HL
	SET_OPCODE 0x2a,LD_HL_nn_o
	SET_OPCODE 0x2b,DEC_HL
	SET_OPCODE 0x2c,INC_L
	SET_OPCODE 0x2d,DEC_L
	SET_OPCODE 0x2e,LD_L_n
	SET_OPCODE 0x2f,CPL
	SET_OPCODE 0x30,JR_NC_e
	SET_OPCODE 0x31,LD_SP_nn
	SET_OPCODE 0x32,LD_nn_A
	SET_OPCODE 0x33,INC_SP
	SET_OPCODE 0x34,INC_HL_
	SET_OPCODE 0x35,DEC_HL_
	SET_OPCODE 0x36,LD_HL_n
	SET_OPCODE 0x37,SCF
	SET_OPCODE 0x38,JR_C_e
	SET_OPCODE 0x39,ADD_HL_SP
	SET_OPCODE 0x3a,LD_A_nn_
	SET_OPCODE 0x3b,DEC_SP
	SET_OPCODE 0x3c,INC_A
	SET_OPCODE 0x3d,DEC_A
	SET_OPCODE 0x3e,LD_A_n
	SET_OPCODE 0x3f,CCF
	SET_OPCODE 0x40,LD_B_B
	SET_OPCODE 0x41,LD_B_C
	SET_OPCODE 0x42,LD_B_D
	SET_OPCODE 0x43,LD_B_E
	SET_OPCODE 0x44,LD_B_H
	SET_OPCODE 0x45,LD_B_L
	SET_OPCODE 0x46,LD_B_HL_
	SET_OPCODE 0x47,LD_B_A
	SET_OPCODE 0x48,LD_C_B
	SET_OPCODE 0x49,LD_C_C
	SET_OPCODE 0x4a,LD_C_D
	SET_OPCODE 0x4b,LD_C_E
	SET_OPCODE 0x4c,LD_C_H
	SET_OPCODE 0x4d,LD_C_L
	SET_OPCODE 0x4e,LD_C_HL_
	SET_OPCODE 0x4f,LD_C_A
	SET_OPCODE 0x50,LD_D_B
	SET_OPCODE 0x51,LD_D_C
	SET_OPCODE 0x52,LD_D_D
	SET_OPCODE 0x53,LD_D_E
	SET_OPCODE 0x54,LD_D_H
	SET_OPCODE 0x55,LD_D_L
	SET_OPCODE 0x56,LD_D_HL_
	SET_OPCODE 0x57,LD_D_A
	SET_OPCODE 0x58,LD_E_B
	SET_OPCODE 0x59,LD_E_C
	SET_OPCODE 0x5a,LD_E_D
	SET_OPCODE 0x5b,LD_E_E
	SET_OPCODE 0x5c,LD_E_H
	SET_OPCODE 0x5d,LD_E_L
	SET_OPCODE 0x5e,LD_E_HL_
	SET_OPCODE 0x5f,LD_E_A
	SET_OPCODE 0x60,LD_H_B
	SET_OPCODE 0x61,LD_H_C
	SET_OPCODE 0x62,LD_H_D
	SET_OPCODE 0x63,LD_H_E
	SET_OPCODE 0x64,LD_H_H
	SET_OPCODE 0x65,LD_H_L
	SET_OPCODE 0x66,LD_H_HL_
	SET_OPCODE 0x67,LD_H_A
	SET_OPCODE 0x68,LD_L_B
	SET_OPCODE 0x69,LD_L_C
	SET_OPCODE 0x6a,LD_L_D
	SET_OPCODE 0x6b,LD_L_E
	SET_OPCODE 0x6c,LD_L_H
	SET_OPCODE 0x6d,LD_L_L
	SET_OPCODE 0x6e,LD_L_HL_
	SET_OPCODE 0x6f,LD_L_A
	SET_OPCODE 0x70,LD_HL_B
	SET_OPCODE 0x71,LD_HL_C
	SET_OPCODE 0x72,LD_HL_D
	SET_OPCODE 0x73,LD_HL_E
	SET_OPCODE 0x74,LD_HL_H
	SET_OPCODE 0x75,LD_HL_L
	SET_OPCODE 0x76,HALT_
	SET_OPCODE 0x77,LD_HL_A
	SET_OPCODE 0x78,LD_A_B
	SET_OPCODE 0x79,LD_A_C
	SET_OPCODE 0x7a,LD_A_D
	SET_OPCODE 0x7b,LD_A_E
	SET_OPCODE 0x7c,LD_A_H
	SET_OPCODE 0x7d,LD_A_L
	SET_OPCODE 0x7e,LD_A_HL_
	SET_OPCODE 0x7f,LD_A_A
	SET_OPCODE 0x80,ADD_A_B
	SET_OPCODE 0x81,ADD_A_C
	SET_OPCODE 0x82,ADD_A_D
	SET_OPCODE 0x83,ADD_A_E
	SET_OPCODE 0x84,ADD_A_H
	SET_OPCODE 0x85,ADD_A_L
	SET_OPCODE 0x86,ADD_A_HL_
	SET_OPCODE 0x87,ADD_A_A
	SET_OPCODE 0x88,ADC_A_B
	SET_OPCODE 0x89,ADC_A_C
	SET_OPCODE 0x8a,ADC_A_D
	SET_OPCODE 0x8b,ADC_A_E
	SET_OPCODE 0x8c,ADC_A_H
	SET_OPCODE 0x8d,ADC_A_L
	SET_OPCODE 0x8e,ADC_A_HL_
	SET_OPCODE 0x8f,ADC_A_A
	SET_OPCODE 0x90,SUB_B
	SET_OPCODE 0x91,SUB_C
	SET_OPCODE 0x92,SUB_D
	SET_OPCODE 0x93,SUB_E
	SET_OPCODE 0x94,SUB_H
	SET_OPCODE 0x95,SUB_L
	SET_OPCODE 0x96,SUB_HL_
	SET_OPCODE 0x97,SUB_A
	SET_OPCODE 0x98,SBC_A_B
	SET_OPCODE 0x99,SBC_A_C
	SET_OPCODE 0x9a,SBC_A_D
	SET_OPCODE 0x9b,SBC_A_E
	SET_OPCODE 0x9c,SBC_A_H
	SET_OPCODE 0x9d,SBC_A_L
	SET_OPCODE 0x9e,SBC_A_HL_
	SET_OPCODE 0x9f,SBC_A_A
	SET_OPCODE 0xa0,AND_B
	SET_OPCODE 0xa1,AND_C
	SET_OPCODE 0xa2,AND_D
	SET_OPCODE 0xa3,AND_E
	SET_OPCODE 0xa4,AND_H
	SET_OPCODE 0xa5,AND_L
	SET_OPCODE 0xa6,AND_HL_
	SET_OPCODE 0xa7,AND_A
	SET_OPCODE 0xa8,XOR_B
	SET_OPCODE 0xa9,XOR_C
	SET_OPCODE 0xaa,XOR_D
	SET_OPCODE 0xab,XOR_E
	SET_OPCODE 0xac,XOR_H
	SET_OPCODE 0xad,XOR_L
	SET_OPCODE 0xae,XOR_HL_
	SET_OPCODE 0xaf,XOR_A
	SET_OPCODE 0xb0,OR_B
	SET_OPCODE 0xb1,OR_C
	SET_OPCODE 0xb2,OR_D
	SET_OPCODE 0xb3,OR_E
	SET_OPCODE 0xb4,OR_H
	SET_OPCODE 0xb5,OR_L
	SET_OPCODE 0xb6,OR_HL_
	SET_OPCODE 0xb7,OR_A
	SET_OPCODE 0xb8,CP_B
	SET_OPCODE 0xb9,CP_C
	SET_OPCODE 0xba,CP_D
	SET_OPCODE 0xbb,CP_E
	SET_OPCODE 0xbc,CP_H
	SET_OPCODE 0xbd,CP_L
	SET_OPCODE 0xbe,CP_HL_
	SET_OPCODE 0xbf,CP_A
	SET_OPCODE 0xc0,RET_NZ
	SET_OPCODE 0xc1,POP_BC
	SET_OPCODE 0xc2,JP_nn_NZ
	SET_OPCODE 0xc3,JMP_nn
	SET_OPCODE 0xc4,CALL_nn_NZ
	SET_OPCODE 0xc5,PUSH_BC
	SET_OPCODE 0xc6,ADD_A_n
	SET_OPCODE 0xc7,RST_0
	SET_OPCODE 0xc8,RET_Z
	SET_OPCODE 0xc9,RET_
	SET_OPCODE 0xca,JP_nn_Z
	SET_OPCODE 0xcb,noinstr     
	SET_OPCODE 0xcc,CALL_nn_Z
	SET_OPCODE 0xcd,CALL_nn
	SET_OPCODE 0xce,ADC_A_n
	SET_OPCODE 0xcf,RST_8
	SET_OPCODE 0xd0,RET_NC
	SET_OPCODE 0xd1,POP_DE
	SET_OPCODE 0xd2,JP_nn_NC
	SET_OPCODE 0xd3,OUT_n_A
	SET_OPCODE 0xd4,CALL_nn_NC
	SET_OPCODE 0xd5,PUSH_DE
	SET_OPCODE 0xd6,SUB_n
	SET_OPCODE 0xd7,RST_10
	SET_OPCODE 0xd8,RET_C
	SET_OPCODE 0xd9,EXX
	SET_OPCODE 0xda,JP_nn_C
	SET_OPCODE 0xdb,IN_A_n_
	SET_OPCODE 0xdc,CALL_nn_C
	SET_OPCODE 0xdd,noinstr   
	SET_OPCODE 0xde,SBC_A_n
	SET_OPCODE 0xdf,RST_18
	SET_OPCODE 0xe0,RET_PO
	SET_OPCODE 0xe1,POP_HL
	SET_OPCODE 0xe2,JP_nn_PO
	SET_OPCODE 0xe3,EX_SP_HL
	SET_OPCODE 0xe4,CALL_nn_PO
	SET_OPCODE 0xe5,PUSH_HL
	SET_OPCODE 0xe6,AND_n
	SET_OPCODE 0xe7,RST_20
	SET_OPCODE 0xe8,RET_PE
	SET_OPCODE 0xe9,JP_HL_
	SET_OPCODE 0xea,JP_nn_PE
	SET_OPCODE 0xeb,EX_DE_HL
	SET_OPCODE 0xec,CALL_nn_PE
	SET_OPCODE 0xed,noinstr   
	SET_OPCODE 0xee,XOR_n
	SET_OPCODE 0xef,RST_28
	SET_OPCODE 0xf0,RET_P
	SET_OPCODE 0xf1,POP_AF
	SET_OPCODE 0xf2,JP_nn_P
	SET_OPCODE 0xf3,DI
	SET_OPCODE 0xf4,CALL_nn_P
	SET_OPCODE 0xf5,PUSH_AF
	SET_OPCODE 0xf6,OR_n
	SET_OPCODE 0xf7,RST_30
	SET_OPCODE 0xf8,RET_M
	SET_OPCODE 0xf9,LD_SP_HL
	SET_OPCODE 0xfa,JP_nn_M
	SET_OPCODE 0xfb,EI
	SET_OPCODE 0xfc,CALL_nn_M
	SET_OPCODE 0xfd,noinstr   
	SET_OPCODE 0xfe,CP_n
	SET_OPCODE 0xff,RST_38
	# opcodeed
	lea opcodeed@END.w(%a5),%a0
	SET_OPCODE 0x40,IN_B_C_
    SET_OPCODE 0x41,OUT_C_B
    SET_OPCODE 0x42,SBC_HL_BC
    SET_OPCODE 0x43,LD_nn_BC
    SET_OPCODE 0x44,NEG_
    SET_OPCODE 0x45,RETN
    SET_OPCODE 0x46,IM_0
    SET_OPCODE 0x47,LD_I_A
    SET_OPCODE 0x48,IN_C_C_
    SET_OPCODE 0x49,OUT_C_C
    SET_OPCODE 0x4a,ADC_HL_BC
    SET_OPCODE 0x4b,LD_BC_nn_
    SET_OPCODE 0x4d,RETI
    SET_OPCODE 0x4f,LD_R_A
    SET_OPCODE 0x50,IN_D_C_
    SET_OPCODE 0x51,OUT_C_D
    SET_OPCODE 0x52,SBC_HL_DE
    SET_OPCODE 0x53,LD_nn_DE
    SET_OPCODE 0x56,IM_1
    SET_OPCODE 0x57,LD_A_I
    SET_OPCODE 0x58,IN_E_C_
    SET_OPCODE 0x59,OUT_C_E
    SET_OPCODE 0x5a,ADC_HL_DE
    SET_OPCODE 0x5b,LD_DE_nn_
    SET_OPCODE 0x5e,IM_2
    SET_OPCODE 0x5f,LD_A_R
    SET_OPCODE 0x60,IN_H_C_
    SET_OPCODE 0x61,OUT_C_H
    SET_OPCODE 0x62,SBC_HL_HL
    SET_OPCODE 0x63,LD_nn_HL
    SET_OPCODE 0x67,RRD
    SET_OPCODE 0x68,IN_L_C_
    SET_OPCODE 0x69,OUT_C_L
    SET_OPCODE 0x6a,ADC_HL_HL
    SET_OPCODE 0x6b,LD_HL_nn_
    SET_OPCODE 0x6f,RLD
    SET_OPCODE 0x72,SBC_HL_SP
    SET_OPCODE 0x73,LD_nn_SP
    SET_OPCODE 0x78,IN_A_C_
    SET_OPCODE 0x79,OUT_C_A
    SET_OPCODE 0x7a,ADC_HL_SP
    SET_OPCODE 0x7b,LD_SP_nn_
    SET_OPCODE 0xa0,LDI
    SET_OPCODE 0xa1,CPI
    SET_OPCODE 0xa2,INI
    SET_OPCODE 0xa3,OUTI
    SET_OPCODE 0xa8,LDD
    SET_OPCODE 0xa9,CPD
    SET_OPCODE 0xaa,IND
    SET_OPCODE 0xab,OUTD
    SET_OPCODE 0xb0,LDIR
    SET_OPCODE 0xb1,CPIR
    SET_OPCODE 0xb2,INIR
    SET_OPCODE 0xb3,OTIR
    SET_OPCODE 0xb8,LDDR
    SET_OPCODE 0xb9,CPDR
    SET_OPCODE 0xba,INDR
    SET_OPCODE 0xbb,OTDR
	# opcodecb
	lea opcodecb@END.w(%a5),%a0
    SET_OPCODE 0x00,RLC_B
    SET_OPCODE 0x01,RLC_C
    SET_OPCODE 0x02,RLC_D
    SET_OPCODE 0x03,RLC_E
    SET_OPCODE 0x04,RLC_H
    SET_OPCODE 0x05,RLC_L
    SET_OPCODE 0x06,RLC_HL_
    SET_OPCODE 0x07,RLC_A
    SET_OPCODE 0x08,RRC_B
    SET_OPCODE 0x09,RRC_C
    SET_OPCODE 0x0a,RRC_D
    SET_OPCODE 0x0b,RRC_E
    SET_OPCODE 0x0c,RRC_H
    SET_OPCODE 0x0d,RRC_L
    SET_OPCODE 0x0e,RRC_HL_
    SET_OPCODE 0x0f,RRC_A
    SET_OPCODE 0x10,RL_B
    SET_OPCODE 0x11,RL_C
    SET_OPCODE 0x12,RL_D
    SET_OPCODE 0x13,RL_E
    SET_OPCODE 0x14,RL_H
    SET_OPCODE 0x15,RL_L
    SET_OPCODE 0x16,RL_HL_
    SET_OPCODE 0x17,RL_A
    SET_OPCODE 0x18,RR_B
    SET_OPCODE 0x19,RR_C
    SET_OPCODE 0x1a,RR_D
    SET_OPCODE 0x1b,RR_E
    SET_OPCODE 0x1c,RR_H
    SET_OPCODE 0x1d,RR_L
    SET_OPCODE 0x1e,RR_HL_
    SET_OPCODE 0x1f,RR_A
    SET_OPCODE 0x20,SLA_B
    SET_OPCODE 0x21,SLA_C
    SET_OPCODE 0x22,SLA_D
    SET_OPCODE 0x23,SLA_E
    SET_OPCODE 0x24,SLA_H
    SET_OPCODE 0x25,SLA_L
    SET_OPCODE 0x26,SLA_HL_
    SET_OPCODE 0x27,SLA_A
    SET_OPCODE 0x28,SRA_B
    SET_OPCODE 0x29,SRA_C
    SET_OPCODE 0x2a,SRA_D
    SET_OPCODE 0x2b,SRA_E
    SET_OPCODE 0x2c,SRA_H
    SET_OPCODE 0x2d,SRA_L
    SET_OPCODE 0x2e,SRA_HL_
    SET_OPCODE 0x2f,SRA_A
    SET_OPCODE 0x38,SRL_B
    SET_OPCODE 0x39,SRL_C
    SET_OPCODE 0x3a,SRL_D
    SET_OPCODE 0x3b,SRL_E
    SET_OPCODE 0x3c,SRL_H
    SET_OPCODE 0x3d,SRL_L
    SET_OPCODE 0x3e,SRL_HL_
    SET_OPCODE 0x3f,SRL_A
    SET_OPCODE 0x40,BIT_0_B
    SET_OPCODE 0x41,BIT_0_C
    SET_OPCODE 0x42,BIT_0_D
    SET_OPCODE 0x43,BIT_0_E
    SET_OPCODE 0x44,BIT_0_H
    SET_OPCODE 0x45,BIT_0_L
    SET_OPCODE 0x46,BIT_0_HL_
    SET_OPCODE 0x47,BIT_0_A
    SET_OPCODE 0x48,BIT_1_B
    SET_OPCODE 0x49,BIT_1_C
    SET_OPCODE 0x4a,BIT_1_D
    SET_OPCODE 0x4b,BIT_1_E
    SET_OPCODE 0x4c,BIT_1_H
    SET_OPCODE 0x4d,BIT_1_L
    SET_OPCODE 0x4e,BIT_1_HL_
    SET_OPCODE 0x4f,BIT_1_A
    SET_OPCODE 0x50,BIT_2_B
    SET_OPCODE 0x51,BIT_2_C
    SET_OPCODE 0x52,BIT_2_D
    SET_OPCODE 0x53,BIT_2_E
    SET_OPCODE 0x54,BIT_2_H
    SET_OPCODE 0x55,BIT_2_L
    SET_OPCODE 0x56,BIT_2_HL_
    SET_OPCODE 0x57,BIT_2_A
    SET_OPCODE 0x58,BIT_3_B
    SET_OPCODE 0x59,BIT_3_C
    SET_OPCODE 0x5a,BIT_3_D
    SET_OPCODE 0x5b,BIT_3_E
    SET_OPCODE 0x5c,BIT_3_H
    SET_OPCODE 0x5d,BIT_3_L
    SET_OPCODE 0x5e,BIT_3_HL_
    SET_OPCODE 0x5f,BIT_3_A
    SET_OPCODE 0x60,BIT_4_B
    SET_OPCODE 0x61,BIT_4_C
    SET_OPCODE 0x62,BIT_4_D
    SET_OPCODE 0x63,BIT_4_E
    SET_OPCODE 0x64,BIT_4_H
    SET_OPCODE 0x65,BIT_4_L
    SET_OPCODE 0x66,BIT_4_HL_
    SET_OPCODE 0x67,BIT_4_A
    SET_OPCODE 0x68,BIT_5_B
    SET_OPCODE 0x69,BIT_5_C
    SET_OPCODE 0x6a,BIT_5_D
    SET_OPCODE 0x6b,BIT_5_E
    SET_OPCODE 0x6c,BIT_5_H
    SET_OPCODE 0x6d,BIT_5_L
    SET_OPCODE 0x6e,BIT_5_HL_
    SET_OPCODE 0x6f,BIT_5_A
    SET_OPCODE 0x70,BIT_6_B
    SET_OPCODE 0x71,BIT_6_C
    SET_OPCODE 0x72,BIT_6_D
    SET_OPCODE 0x73,BIT_6_E
    SET_OPCODE 0x74,BIT_6_H
    SET_OPCODE 0x75,BIT_6_L
    SET_OPCODE 0x76,BIT_6_HL_
    SET_OPCODE 0x77,BIT_6_A
    SET_OPCODE 0x78,BIT_7_B
    SET_OPCODE 0x79,BIT_7_C
    SET_OPCODE 0x7a,BIT_7_D
    SET_OPCODE 0x7b,BIT_7_E
    SET_OPCODE 0x7c,BIT_7_H
    SET_OPCODE 0x7d,BIT_7_L
    SET_OPCODE 0x7e,BIT_7_HL_
    SET_OPCODE 0x7f,BIT_7_A
    SET_OPCODE 0x80,RES_0_B
    SET_OPCODE 0x81,RES_0_C
    SET_OPCODE 0x82,RES_0_D
    SET_OPCODE 0x83,RES_0_E
    SET_OPCODE 0x84,RES_0_H
    SET_OPCODE 0x85,RES_0_L
    SET_OPCODE 0x86,RES_0_HL_
    SET_OPCODE 0x87,RES_0_A
    SET_OPCODE 0x88,RES_1_B
    SET_OPCODE 0x89,RES_1_C
    SET_OPCODE 0x8a,RES_1_D
    SET_OPCODE 0x8b,RES_1_E
    SET_OPCODE 0x8c,RES_1_H
    SET_OPCODE 0x8d,RES_1_L
    SET_OPCODE 0x8e,RES_1_HL_
    SET_OPCODE 0x8f,RES_1_A
    SET_OPCODE 0x90,RES_2_B
    SET_OPCODE 0x91,RES_2_C
    SET_OPCODE 0x92,RES_2_D
    SET_OPCODE 0x93,RES_2_E
    SET_OPCODE 0x94,RES_2_H
    SET_OPCODE 0x95,RES_2_L
    SET_OPCODE 0x96,RES_2_HL_
    SET_OPCODE 0x97,RES_2_A
    SET_OPCODE 0x98,RES_3_B
    SET_OPCODE 0x99,RES_3_C
    SET_OPCODE 0x9a,RES_3_D
    SET_OPCODE 0x9b,RES_3_E
    SET_OPCODE 0x9c,RES_3_H
    SET_OPCODE 0x9d,RES_3_L
    SET_OPCODE 0x9e,RES_3_HL_
    SET_OPCODE 0x9f,RES_3_A
    SET_OPCODE 0xa0,RES_4_B
    SET_OPCODE 0xa1,RES_4_C
    SET_OPCODE 0xa2,RES_4_D
    SET_OPCODE 0xa3,RES_4_E
    SET_OPCODE 0xa4,RES_4_H
    SET_OPCODE 0xa5,RES_4_L
    SET_OPCODE 0xa6,RES_4_HL_
    SET_OPCODE 0xa7,RES_4_A
    SET_OPCODE 0xa8,RES_5_B
    SET_OPCODE 0xa9,RES_5_C
    SET_OPCODE 0xaa,RES_5_D
    SET_OPCODE 0xab,RES_5_E
    SET_OPCODE 0xac,RES_5_H
    SET_OPCODE 0xad,RES_5_L
    SET_OPCODE 0xae,RES_5_HL_
    SET_OPCODE 0xaf,RES_5_A
    SET_OPCODE 0xb0,RES_6_B
    SET_OPCODE 0xb1,RES_6_C
    SET_OPCODE 0xb2,RES_6_D
    SET_OPCODE 0xb3,RES_6_E
    SET_OPCODE 0xb4,RES_6_H
    SET_OPCODE 0xb5,RES_6_L
    SET_OPCODE 0xb6,RES_6_HL_
    SET_OPCODE 0xb7,RES_6_A
    SET_OPCODE 0xb8,RES_7_B
    SET_OPCODE 0xb9,RES_7_C
    SET_OPCODE 0xba,RES_7_D
    SET_OPCODE 0xbb,RES_7_E
    SET_OPCODE 0xbc,RES_7_H
    SET_OPCODE 0xbd,RES_7_L
    SET_OPCODE 0xbe,RES_7_HL_
    SET_OPCODE 0xbf,RES_7_A
    SET_OPCODE 0xc0,SET_0_B
    SET_OPCODE 0xc1,SET_0_C
    SET_OPCODE 0xc2,SET_0_D
    SET_OPCODE 0xc3,SET_0_E
    SET_OPCODE 0xc4,SET_0_H
    SET_OPCODE 0xc5,SET_0_L
    SET_OPCODE 0xc6,SET_0_HL_
    SET_OPCODE 0xc7,SET_0_A
    SET_OPCODE 0xc8,SET_1_B
    SET_OPCODE 0xc9,SET_1_C
    SET_OPCODE 0xca,SET_1_D
    SET_OPCODE 0xcb,SET_1_E
    SET_OPCODE 0xcc,SET_1_H
    SET_OPCODE 0xcd,SET_1_L
    SET_OPCODE 0xce,SET_1_HL_
    SET_OPCODE 0xcf,SET_1_A
    SET_OPCODE 0xd0,SET_2_B
    SET_OPCODE 0xd1,SET_2_C
    SET_OPCODE 0xd2,SET_2_D
    SET_OPCODE 0xd3,SET_2_E
    SET_OPCODE 0xd4,SET_2_H
    SET_OPCODE 0xd5,SET_2_L
    SET_OPCODE 0xd6,SET_2_HL_
    SET_OPCODE 0xd7,SET_2_A
    SET_OPCODE 0xd8,SET_3_B
    SET_OPCODE 0xd9,SET_3_C
    SET_OPCODE 0xda,SET_3_D
    SET_OPCODE 0xdb,SET_3_E
    SET_OPCODE 0xdc,SET_3_H
    SET_OPCODE 0xdd,SET_3_L
    SET_OPCODE 0xde,SET_3_HL_
    SET_OPCODE 0xdf,SET_3_A
    SET_OPCODE 0xe0,SET_4_B
    SET_OPCODE 0xe1,SET_4_C
    SET_OPCODE 0xe2,SET_4_D
    SET_OPCODE 0xe3,SET_4_E
    SET_OPCODE 0xe4,SET_4_H
    SET_OPCODE 0xe5,SET_4_L
    SET_OPCODE 0xe6,SET_4_HL_
    SET_OPCODE 0xe7,SET_4_A
    SET_OPCODE 0xe8,SET_5_B
    SET_OPCODE 0xe9,SET_5_C
    SET_OPCODE 0xea,SET_5_D
    SET_OPCODE 0xeb,SET_5_E
    SET_OPCODE 0xec,SET_5_H
    SET_OPCODE 0xed,SET_5_L
    SET_OPCODE 0xee,SET_5_HL_
    SET_OPCODE 0xef,SET_5_A
    SET_OPCODE 0xf0,SET_6_B
    SET_OPCODE 0xf1,SET_6_C
    SET_OPCODE 0xf2,SET_6_D
    SET_OPCODE 0xf3,SET_6_E
    SET_OPCODE 0xf4,SET_6_H
    SET_OPCODE 0xf5,SET_6_L
    SET_OPCODE 0xf6,SET_6_HL_
    SET_OPCODE 0xf7,SET_6_A
    SET_OPCODE 0xf8,SET_7_B
    SET_OPCODE 0xf9,SET_7_C
    SET_OPCODE 0xfa,SET_7_D
    SET_OPCODE 0xfb,SET_7_E
    SET_OPCODE 0xfc,SET_7_H
    SET_OPCODE 0xfd,SET_7_L
    SET_OPCODE 0xfe,SET_7_HL_
    SET_OPCODE 0xff,SET_7_A
	# opcodedd
	lea opcodedd@END.w(%a5),%a0
    SET_OPCODE 0x09,ADD_IX_BC
    SET_OPCODE 0x19,ADD_IX_DE
    SET_OPCODE 0x21,LD_IX_nn
    SET_OPCODE 0x22,LD_nn_IX
    SET_OPCODE 0x23,INC_IX
	SET_OPCODE 0x26,LD_IXh_n
    SET_OPCODE 0x29,ADD_IX_IX
    SET_OPCODE 0x2a,LD_IX_nn_
    SET_OPCODE 0x2b,DEC_IX
	SET_OPCODE 0x2e,LD_IXl_n
    SET_OPCODE 0x34,INC_IX_d
    SET_OPCODE 0x35,DEC_IX_d
    SET_OPCODE 0x36,LD_IX_d_n
    SET_OPCODE 0x39,ADD_IX_SP
    SET_OPCODE 0x46,LD_B_IX_d
    SET_OPCODE 0x4e,LD_C_IX_d
    SET_OPCODE 0x56,LD_D_IX_d
    SET_OPCODE 0x5e,LD_E_IX_d
    SET_OPCODE 0x66,LD_H_IX_d
	SET_OPCODE 0x67,LD_IXh_A
    SET_OPCODE 0x6e,LD_L_IX_d
	SET_OPCODE 0x6f,LD_IXl_A
    SET_OPCODE 0x70,LD_IX_d_B
    SET_OPCODE 0x71,LD_IX_d_C
    SET_OPCODE 0x72,LD_IX_d_D
    SET_OPCODE 0x73,LD_IX_d_E
    SET_OPCODE 0x74,LD_IX_d_H
    SET_OPCODE 0x75,LD_IX_d_L
    SET_OPCODE 0x77,LD_IX_d_A
    SET_OPCODE 0x7e,LD_A_IX_d
    SET_OPCODE 0x86,ADD_A_IX_d
    SET_OPCODE 0x8e,ADC_A_IX_d
    SET_OPCODE 0x96,SUB_IX_d
    SET_OPCODE 0x9e,SBC_A_IX_d
    SET_OPCODE 0xa6,AND_IX_d
    SET_OPCODE 0xae,XOR_IX_d
    SET_OPCODE 0xb6,OR_IX_d
    SET_OPCODE 0xbe,CP_IX_d
    SET_OPCODE 0xcb,noinstr
    SET_OPCODE 0xe1,POP_IX
    SET_OPCODE 0xe3,EX_SP_IX
    SET_OPCODE 0xe5,PUSH_IX
    SET_OPCODE 0xe9,JP_IX_
    SET_OPCODE 0xf9,LD_SP_IX
	# opcodefd
	lea opcodefd@END.w(%a5),%a0	
    SET_OPCODE 0x09,ADD_IY_BC
    SET_OPCODE 0x19,ADD_IY_DE
    SET_OPCODE 0x21,LD_IY_nn
    SET_OPCODE 0x22,LD_nn_IY
    SET_OPCODE 0x23,INC_IY
	SET_OPCODE 0x26,LD_IYh_n
    SET_OPCODE 0x29,ADD_IY_IY
    SET_OPCODE 0x2a,LD_IY_nn_
    SET_OPCODE 0x2b,DEC_IY
	SET_OPCODE 0x2e,LD_IYl_n
    SET_OPCODE 0x34,INC_IY_d
    SET_OPCODE 0x35,DEC_IY_d
    SET_OPCODE 0x36,LD_IY_d_n
    SET_OPCODE 0x39,ADD_IY_SP
    SET_OPCODE 0x46,LD_B_IY_d
    SET_OPCODE 0x4e,LD_C_IY_d
    SET_OPCODE 0x56,LD_D_IY_d
    SET_OPCODE 0x5e,LD_E_IY_d
    SET_OPCODE 0x66,LD_H_IY_d
	SET_OPCODE 0x67,LD_IYh_A
    SET_OPCODE 0x6e,LD_L_IY_d
	SET_OPCODE 0x6f,LD_IYl_A
    SET_OPCODE 0x70,LD_IY_d_B
    SET_OPCODE 0x71,LD_IY_d_C
    SET_OPCODE 0x72,LD_IY_d_D
    SET_OPCODE 0x73,LD_IY_d_E
    SET_OPCODE 0x74,LD_IY_d_H
    SET_OPCODE 0x75,LD_IY_d_L
    SET_OPCODE 0x77,LD_IY_d_A
    SET_OPCODE 0x7e,LD_A_IY_d
    SET_OPCODE 0x86,ADD_A_IY_d
    SET_OPCODE 0x8e,ADC_A_IY_d
    SET_OPCODE 0x96,SUB_IY_d
    SET_OPCODE 0x9e,SBC_A_IY_d
    SET_OPCODE 0xa6,AND_IY_d
    SET_OPCODE 0xae,XOR_IY_d
    SET_OPCODE 0xb6,OR_IY_d
    SET_OPCODE 0xbe,CP_IY_d
    SET_OPCODE 0xcb,noinstr
    SET_OPCODE 0xe1,POP_IY
    SET_OPCODE 0xe3,EX_SP_IY
    SET_OPCODE 0xe5,PUSH_IY
    SET_OPCODE 0xe9,JP_IY_
    SET_OPCODE 0xf9,LD_SP_IY
	# opcodeddcb
	lea opcodeddcb@END.w(%a5),%a0
    SET_OPCODE 0x06,RLC_IX_d
    SET_OPCODE 0x0e,RRC_IX_d
    SET_OPCODE 0x16,RL_IX_d
    SET_OPCODE 0x1e,RR_IX_d
    SET_OPCODE 0x26,SLA_IX_d
    SET_OPCODE 0x2e,SRA_IX_d
    SET_OPCODE 0x3e,SRL_IX_d
    SET_OPCODE 0x46,BIT_0_IX_d
    SET_OPCODE 0x4e,BIT_1_IX_d
    SET_OPCODE 0x56,BIT_2_IX_d
    SET_OPCODE 0x5e,BIT_3_IX_d
    SET_OPCODE 0x66,BIT_4_IX_d
    SET_OPCODE 0x6e,BIT_5_IX_d
    SET_OPCODE 0x76,BIT_6_IX_d
    SET_OPCODE 0x7e,BIT_7_IX_d
    SET_OPCODE 0x86,RES_0_IX_d
    SET_OPCODE 0x8e,RES_1_IX_d
    SET_OPCODE 0x96,RES_2_IX_d
    SET_OPCODE 0x9e,RES_3_IX_d
    SET_OPCODE 0xa6,RES_4_IX_d
    SET_OPCODE 0xae,RES_5_IX_d
    SET_OPCODE 0xb6,RES_6_IX_d
    SET_OPCODE 0xbe,RES_7_IX_d
    SET_OPCODE 0xc6,SET_0_IX_d
    SET_OPCODE 0xce,SET_1_IX_d
    SET_OPCODE 0xd6,SET_2_IX_d
    SET_OPCODE 0xde,SET_3_IX_d
    SET_OPCODE 0xe6,SET_4_IX_d
    SET_OPCODE 0xee,SET_5_IX_d
    SET_OPCODE 0xf6,SET_6_IX_d
    SET_OPCODE 0xfe,SET_7_IX_d
	# opcodefdcb
	lea opcodefdcb@END.w(%a5),%a0
    SET_OPCODE 0x06,RLC_IY_d
    SET_OPCODE 0x0e,RRC_IY_d
    SET_OPCODE 0x16,RL_IY_d
    SET_OPCODE 0x1e,RR_IY_d
    SET_OPCODE 0x26,SLA_IY_d
    SET_OPCODE 0x2e,SRA_IY_d
    SET_OPCODE 0x3e,SRL_IY_d
    SET_OPCODE 0x46,BIT_0_IY_d
    SET_OPCODE 0x4e,BIT_1_IY_d
    SET_OPCODE 0x56,BIT_2_IY_d
    SET_OPCODE 0x5e,BIT_3_IY_d
    SET_OPCODE 0x66,BIT_4_IY_d
    SET_OPCODE 0x6e,BIT_5_IY_d
    SET_OPCODE 0x76,BIT_6_IY_d
    SET_OPCODE 0x7e,BIT_7_IY_d
    SET_OPCODE 0x86,RES_0_IY_d
    SET_OPCODE 0x8e,RES_1_IY_d
    SET_OPCODE 0x96,RES_2_IY_d
    SET_OPCODE 0x9e,RES_3_IY_d
    SET_OPCODE 0xa6,RES_4_IY_d
    SET_OPCODE 0xae,RES_5_IY_d
    SET_OPCODE 0xb6,RES_6_IY_d
    SET_OPCODE 0xbe,RES_7_IY_d
    SET_OPCODE 0xc6,SET_0_IY_d
    SET_OPCODE 0xce,SET_1_IY_d
    SET_OPCODE 0xd6,SET_2_IY_d
    SET_OPCODE 0xde,SET_3_IY_d
    SET_OPCODE 0xe6,SET_4_IY_d
    SET_OPCODE 0xee,SET_5_IY_d
    SET_OPCODE 0xf6,SET_6_IY_d
    SET_OPCODE 0xfe,SET_7_IY_d
	RESET_REGS
	# MEM_ALLOC 0xc000,%a1
	GETL memram,%a1
	SETL MEM,%a1
	# MEM_ALLOC 0x4000,%a1
	GETL memrom,%a1
	SETL ROM,%a1
	lea t_rom_file(%pc),%a2
	FILEDUMP %a1,%a2,0,0x4000

	POPALL
	unlk %a6
	rts

	.even
t_rom_file:
	.ascii "SpectrumRom\0"

.macro LSNAB var
	move.b (%a1),%d0
	SETB \var,%d0
	add.l #1,%a1
	.endm

.macro LSNABR var
	move.b (%a1),%d0
	move.b %d0,\var
	add.l #1,%a1
	.endm

.macro LSNAW var
	move.b 1(%a1),%d0
	lsl.w #8,%d0
	move.b (%a1),%d0
	SETW \var,%d0
	add.l #2,%a1
	.endm

	.even
	.globl	LoadSNA
LoadSNA:
	# UCHAR LoadSNA (char *file)
	link.w %a6,#0
	PUSHALL
	GETL MEM,%a1
	move.l 8(%a6),%d1
	move.l %a1,-(%sp)
	FILEDUMP %a1,%d1,0,27
	move.l (%sp)+,%a1
	LSNAB REG_I
	LSNAW REG_HL_
	LSNAW REG_DE_
	LSNAW REG_BC_
	LSNAW REG_AF_
	LSNAB REG_L
	LSNAB REG_H
	LSNAB REG_E
	LSNAB REG_D
	LSNAB REG_C
	LSNAB REG_B
	LSNAW REG_IY
	LSNAW REG_IX
	clr.b %d3
	LSNABR %d0
	cmp.b #0,%d0
	beq load1
	move.b #1,%d0
load1:
	move.b %d0,%d1
	SET_IFF1 %d0
	SET_IFF2 %d1
	LSNAB REG_R
	LSNAB REG_F
	LSNAB REG_A
	LSNAW REG_SP
	LSNABR %d0
	SET_IM %d0
	SAVE_R FLAGS,%d3
	# border
	GETL MEM,%a1
	move.l 8(%a6),%d1
	FILEDUMP %a1,%d1,27,0xc000
	# RETN
	LOAD_REGS
	
	S_F
	bclr.b #3,%d1
	bclr.b #5,%d1
	E_F
	
	lea SaveScreen(%pc),%a1
	jsr (%a1)

	PCONTW0 SP
	SET_PC %d0
	ADD_SP #2
	SAVE_REGS
	POPALL
	unlk %a6
	rts

.macro SSNAB var
	GETB \var,%d0
	move.b %d0,(%a1)
	add.l #1,%a1
	.endm

.macro SSNABR var
	move.b \var,%d0
	move.b %d0,(%a1)
	add.l #1,%a1
	.endm

.macro SSNAW var
	GETW \var,%d0
	move.b %d0,(%a1)
	lsr.w #8,%d0
	move.b %d0,1(%a1)
	add.l #2,%a1
	.endm

	.even
	.globl	SaveSNA
SaveSNA:
	# UCHAR SaveSNA (char *file)
	link.w %a6,#0
	PUSHALL
	move.l 8(%a6),%d1
	FILECREATE %d1,27+0xc000
	LOAD_REGS
	SUB_SP #2
	GET_SP %a0
	GET_PC %d0
	STOREW
	SAVE_REGS
	lea SNA@END.w(%a5),%a1
	SSNAB REG_I
	SSNAW REG_HL_
	SSNAW REG_DE_
	SSNAW REG_BC_
	SSNAW REG_AF_
	SSNAB REG_L
	SSNAB REG_H
	SSNAB REG_E
	SSNAB REG_D
	SSNAB REG_C
	SSNAB REG_B
	SSNAW REG_IY
	SSNAW REG_IX
	GET_IFF1 %d0
	SSNABR %d0
	SSNAB REG_R
	SSNAB REG_F
	SSNAB REG_A
	SSNAW REG_SP
	GET_IM %d0
	SSNABR %d0
	# border
	lea SNA@END.w(%a5),%a1
	move.l 8(%a6),%d1
	FILEWRITE %d1,%a1,0,27
	move.l 8(%a6),%d1
	GETL MEM,%a1
	FILEWRITE %d1,%a1,27,0xc000
	POPALL
	unlk %a6	
	rts

	.even
	.globl	End
End:	
	link.w %a6,#0
	/*
	PUSHALL
	GETL MEM,%a1
	MEM_FREE %a1
	GETL ROM,%a1
	MEM_FREE %a1
	POPALL
	*/
	unlk %a6
	rts

	.even
	.globl	InDebug
InDebug:
	move.b #0,%d0
	rts

	.even
	.globl	DoDebug
DoDebug:
	rts

	.even
	.globl	Refresh
Refresh:
	link.w %a6,#0
	PUSHALL
	GETL MEM,%a4
	move.l #0x4000,%a0
	clr.l %d7
	move.b #0xff,%d1
refnext:
	move.b (%a4,%d7.l),%d0
	lea screenmem(%pc),%a1
	jsr (%a1)
	add.w #1,%d7
	add.w #1,%a0
	cmp.w #0x1800,%d7
	bcs refnext

	lea SaveScreen(%pc),%a0
	jsr (%a0)

	POPALL
	unlk %a6
	rts

	.even
SaveScreen:
	lea sb@END.w(%a5),%a0
	clr.l %d7
save1:
	move.b (%a4,%d7.l),%d0
	move.b %d0,(%a0,%d7.l)
	add.w #1,%d7
	cmp.w #0x1b00,%d7
	bcs save1
	rts

	.even
RefreshInt:
	move.l %d1,-(%sp)
	move.l %d2,-(%sp)
	move.l %d3,-(%sp)
	lea sb@END.w(%a5),%a2
	move.l #0x4000,%a0
	clr.l %d7
refint:
	move.b (%a4,%d7.l),%d0
	move.b (%a2,%d7.l),%d2
	cmp.b %d2,%d0
	beq nrefin
	# m = ~((o&c)|((~o)&(~c)));
	move.b %d0,%d1
	move.b %d2,%d3
	not.b %d1
	not.b %d3
	and.b %d1,%d3
	move.b %d0,%d1
	and.b %d2,%d1
	or.b %d3,%d1
	not.b %d1
	move.b %d0,(%a2,%d7.l)
	lea screenmem(%pc),%a1
	jsr (%a1)
nrefin:
	add.w #1,%d7
	add.w #1,%a0
	cmp.w #0x1800,%d7
	bcs refint
refintc:
	move.b (%a4,%d7.l),%d0
	cmp.b (%a2,%d7.l),%d0
	beq nrefinc
	move.b %d0,(%a2,%d7.l)
	lea colormem(%pc),%a1
	jsr (%a1)
nrefinc:
	add.w #1,%d7
	add.w #1,%a0
	cmp.w #0x1b00,%d7
	bcs refintc	
	move.l (%sp)+,%d3
	move.l (%sp)+,%d2
	move.l (%sp)+,%d1
	rts


# debug
	.even
t_code:
	.ascii "CODE: %04Lx\0"
	.even
t_af:
	.ascii "AF: %04Lx\0"
	.even
t_bc:
	.ascii "BC: %04Lx\0"
	.even
t_de:
	.ascii "DE: %04Lx\0"
	.even
t_hl:
	.ascii "HL: %04Lx\0"
	.even
t_ix:
	.ascii "IX: %04Lx\0"
	.even
t_iy:
	.ascii "IY: %04Lx\0"
	.even
t_pc:
	.ascii "PC: %04Lx\0"
	.even
t_sp:
	.ascii "SP: %04Lx\0"
	.even
t_ir:
	.ascii "IR: %04Lx\0"
# debug

	.even
	.globl	Execute
Execute:
	#void Execute (ULONG time)
	link.w %a6,#0
	PUSHALL
	LOAD_REGS
	SETL TIME,#0
newcode:
	PCONTB0 PC
	/*
	# debug
	PUSHALL
	move.l %d0,%d7
	GET_PC %d0
	cmp.w #0x11ea,%d0
	bne nodeb3
	SETB trace,#1
nodeb3:
	GETB trace,%d0
	beq nodeb2
	move.l %d7,%d0
	PRINT t_code,%d0,#0,#0
	PRINTLN
nodeb2:
	POPALL	
	# debug
	*/
	cmp.b #0xed,%d0
	bne c_cb
	PCONTB PC,#1
	SET_INCR #2
	EXEC_CODE opcodeed,%d0
c_cb:
	cmp.b #0xcb,%d0
	bne c_dd
	PCONTB PC,#1
	SET_INCR #2
	EXEC_CODE opcodecb,%d0
c_dd:
	cmp.b #0xdd,%d0
	bne c_fd
	PCONTB PC,#1
	cmp.b #0xcb,%d0
	bne c_dd2
	PCONTB PC,#3
	SET_INCR #2
	EXEC_CODE opcodeddcb,%d0
c_dd2:
	SET_INCR #2
	EXEC_CODE opcodedd,%d0
c_fd:
	cmp.b #0xfd,%d0
	bne c_nor
	PCONTB PC,#1
	cmp.b #0xcb,%d0
	bne c_fd2	
	PCONTB PC,#3
	SET_INCR #2
	EXEC_CODE opcodefdcb,%d0
c_fd2:
	SET_INCR #2
	EXEC_CODE opcodefd,%d0
c_nor:
	SET_INCR #1    
	EXEC_CODE opcode,%d0
nextins:
/*
	GETL temp,%a0
	cmp.l %a0,%a1
	beq ok
	BREAK
ok:
*/
	GET_NEI %d0
	cmp.b #0,%d0
	beq nonei
	add.b #1,%d0
	cmp.b #3,%d0
	bcs neion
	clr.b %d0
	SET_NEI %d0
	move.b #1,%d0
	SET_IFF1 %d0
	move.b #1,%d0
	SET_IFF2 %d0
	bra nonei
	.even
neion:
	SET_NEI %d0    
nonei:
	GET_INCR %d0
	S_R
	btst.b #7,%d3
	bne ron
	add.b %d0,%d3
	bclr.b #7,%d3
	bra roff
	.even
ron:
	add.b %d0,%d3
	bset.b #7,%d3
roff:
	E_R
	/*
	# debug
	GETB trace,%d0
	cmp.b #0,%d0
	beq nodeb1
	PUSHALL
	FLAGS_TO_F
	clr.l %d0
	GET_AF %d0
	PRINT t_af,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_BC %d0
	PRINT t_bc,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_DE %d0
	PRINT t_de,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_HL %d0
	PRINT t_hl,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_IX %d0
	PRINT t_ix,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_IY %d0
	PRINT t_iy,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_PC %d0
	PRINT t_pc,%d0,#0,#0
	PRINTLN
	POPALL
	PUSHALL
	clr.l %d0
	GET_SP %d0
	PRINT t_sp,%d0,#0,#0	
	PRINTLN
	POPALL	
	PUSHALL
	clr.l %d0
	GET_IR %d0
	PRINT t_ir,%d0,#0,#0	
	PRINTLN
	POPALL	
	BREAK
nodeb1:
	# debug
	*/
	GET_time %d0
	cmp.l #5000,%d0
	bcs newcode
	SUML TIME,%d0
	SUML TIME_INT,%d0
	SUML TIME_FLASH,%d0
	GETL TIME_INT,%d0
	cmp.l #20000,%d0
	bcs noint
	RESL TIME_INT,#20000 
	GET_IFF1 %d0
	cmp.b #0,%d0
	bne inter
	GET_IM %d0
	cmp.b #0,%d0
	beq inter
	bra noint
inter:
	clr.b %d0
	SET_IFF1 %d0
	GET_IM %d0
	cmp.b #0,%d0
	beq int0
	cmp.b #1,%d0
	beq int1
	cmp.b #2,%d0
	beq int2
	# default
	bra noint
int0:
int1:
	S_SP
	sub.w #2,%d4
	move.w %d4,%a0
	E_SP
	S_PC
	move.w %d4,%d0
	STOREW
	move.w #0x0038,%d4
	E_PC
	bra noint
int2:
	SUB_SP #2
	GET_SP %a0
	GET_PC %d0
	STOREW
	GET_I %d0
	lsl.w #8,%d0
	or.w #0x00ff,%d0
	SET_PC %d0
	PCONTW0 PC
	SET_PC %d0
noint:
	GETL TIME_FLASH,%d0
	cmp.l #250000,%d0
	bcs exit
	RESL TIME_FLASH,#250000
	GET_flash %d0
	bchg.b #0,%d0
	SET_flash %d0

	move.w #0x5800,%d7
flash1:
	move.w %d7,%a0
	READB0
	and.b #0x80,%d0
	beq flash2
	move.w %d7,%a0
	lea colormem(%pc),%a1
	jsr (%a1)
flash2:
	add.w #1,%d7
	cmp.w #0x5b00,%d7
	bcs flash1

exit:
	GETB rsc,%d0
	add.b #1,%d0
	cmp.b #2,%d0
	bcs norate
	lea RefreshInt(%pc),%a1
	jsr (%a1)
	clr.l %d0
norate:
	SETB rsc,%d0

	SET_time #0
	GETL TIME,%d0
	cmp.l 8(%a6),%d0
	bcs newcode
	#bra newcode
	SAVE_REGS
	POPALL
	unlk %a6	
	rts

.macro TESTKEY k,num,val
	cmp.b #\k,%d0
	bne key\@
	ANDB KEYBOARD\num,#\val
	bra lkey
key\@:
	.endm

	.even
	.globl	SetKey
SetKey:
	#void SetKey (UCHAR k)
	link.w %a6,#0
	PUSHALL
	move.b 8(%a6),%d0
	TESTKEY ',',0,0xfe
	TESTKEY 'z',0,0xfd
	TESTKEY 'x',0,0xfb
	TESTKEY 'c',0,0xf7
	TESTKEY 'v',0,0xef
	TESTKEY 'a',1,0xfe
	TESTKEY 's',1,0xfd
	TESTKEY 'd',1,0xfb
	TESTKEY 'f',1,0xf7
	TESTKEY 'g',1,0xef
	TESTKEY 'q',2,0xfe
	TESTKEY 'w',2,0xfd
	TESTKEY 'e',2,0xfb
	TESTKEY 'r',2,0xf7
	TESTKEY 't',2,0xef
	TESTKEY '1',3,0xfe
	TESTKEY '2',3,0xfd
	TESTKEY '3',3,0xfb
	TESTKEY '4',3,0xf7
	TESTKEY '5',3,0xef
	TESTKEY '0',4,0xfe
	TESTKEY '9',4,0xfd
	TESTKEY '8',4,0xfb
	TESTKEY '7',4,0xf7
	TESTKEY '6',4,0xef
	TESTKEY 'p',5,0xfe
	TESTKEY 'o',5,0xfd
	TESTKEY 'i',5,0xfb
	TESTKEY 'u',5,0xf7
	TESTKEY 'y',5,0xef
	TESTKEY '\n',6,0xfe
	TESTKEY '\r',6,0xfe
	TESTKEY 'l',6,0xfd
	TESTKEY 'k',6,0xfb
	TESTKEY 'j',6,0xf7
	TESTKEY 'h',6,0xef
	TESTKEY ' ',7,0xfe
	TESTKEY '.',7,0xfd
	TESTKEY 'm',7,0xfb
	TESTKEY 'n',7,0xf7
	TESTKEY 'b',7,0xef
lkey:
	POPALL
	unlk %a6
	rts

	.even
	.globl	SetKempston
SetKempston:
	#void SetKempston (UCHAR k)
	link.w %a6,#0
	PUSHALL
	move.b 8(%a6),%d0
	SETB KEMPSTON,%d0
	POPALL
	unlk %a6
	rts

.macro RESETKEY k,num,val
	cmp.b #\k,%d0
	bne key\@
	ORB KEYBOARD\num,#\val
	bra rkey
key\@:
	.endm

	.even
	.globl	ResetKey
ResetKey:
	#void ResetKey (UCHAR k)
	link.w %a6,#0
	PUSHALL
	move.b 8(%a6),%d0
	RESETKEY ',',0,0x01
	RESETKEY 'z',0,0x02
	RESETKEY 'x',0,0x04
	RESETKEY 'c',0,0x08
	RESETKEY 'v',0,0x10
	RESETKEY 'a',1,0x01
	RESETKEY 's',1,0x02
	RESETKEY 'd',1,0x04
	RESETKEY 'f',1,0x08
	RESETKEY 'g',1,0x10
	RESETKEY 'q',2,0x01
	RESETKEY 'w',2,0x02
	RESETKEY 'e',2,0x04
	RESETKEY 'r',2,0x08
	RESETKEY 't',2,0x10
	RESETKEY '1',3,0x01
	RESETKEY '2',3,0x02
	RESETKEY '3',3,0x04
	RESETKEY '4',3,0x08
	RESETKEY '5',3,0x10
	RESETKEY '0',4,0x01
	RESETKEY '9',4,0x02
	RESETKEY '8',4,0x04
	RESETKEY '7',4,0x08
	RESETKEY '6',4,0x10
	RESETKEY 'p',5,0x01
	RESETKEY 'o',5,0x02
	RESETKEY 'i',5,0x04
	RESETKEY 'u',5,0x08
	RESETKEY 'y',5,0x10
	RESETKEY '\n',6,0x01
	RESETKEY '\r',6,0x01
	RESETKEY 'l',6,0x02
	RESETKEY 'k',6,0x04
	RESETKEY 'j',6,0x08
	RESETKEY 'h',6,0x10
	RESETKEY ' ',7,0x01
	RESETKEY '.',7,0x02
	RESETKEY 'm',7,0x04
	RESETKEY 'n',7,0x08
	RESETKEY 'b',7,0x10
rkey:
	POPALL
	unlk %a6
	rts

# --------------------------------------------------------------------------------------

F_RES_b_r 0,C,%d1
F_RES_b_r 1,C,%d1
F_RES_b_r 2,C,%d1
F_RES_b_r 3,C,%d1
F_RES_b_r 4,C,%d1
F_RES_b_r 5,C,%d1
F_RES_b_r 6,C,%d1
F_RES_b_r 7,C,%d1

F_RES_b_r 0,D,%d2
F_RES_b_r 1,D,%d2
F_RES_b_r 2,D,%d2
F_RES_b_r 3,D,%d2
F_RES_b_r 4,D,%d2
F_RES_b_r 5,D,%d2
F_RES_b_r 6,D,%d2
F_RES_b_r 7,D,%d2

F_RES_b_r 0,E,%d2
F_RES_b_r 1,E,%d2
F_RES_b_r 2,E,%d2
F_RES_b_r 3,E,%d2
F_RES_b_r 4,E,%d2
F_RES_b_r 5,E,%d2
F_RES_b_r 6,E,%d2
F_RES_b_r 7,E,%d2

F_RES_b_r 0,H,%d2
F_RES_b_r 1,H,%d2
F_RES_b_r 2,H,%d2
F_RES_b_r 3,H,%d2
F_RES_b_r 4,H,%d2
F_RES_b_r 5,H,%d2
F_RES_b_r 6,H,%d2
F_RES_b_r 7,H,%d2

F_RES_b_r 0,L,%d2
F_RES_b_r 1,L,%d2
F_RES_b_r 2,L,%d2
F_RES_b_r 3,L,%d2
F_RES_b_r 4,L,%d2
F_RES_b_r 5,L,%d2
F_RES_b_r 6,L,%d2
F_RES_b_r 7,L,%d2

F_RES_b_r 0,A,%d1
F_RES_b_r 1,A,%d1
F_RES_b_r 2,A,%d1
F_RES_b_r 3,A,%d1
F_RES_b_r 4,A,%d1
F_RES_b_r 5,A,%d1
F_RES_b_r 6,A,%d1
F_RES_b_r 7,A,%d1

FUNC RLC_HL_
	PCONTB0 HL
	rol.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_RLC_r l
FUNC RLC_\l
	GET_\l %d0
	rol.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_RLC_r B
F_RLC_r C
F_RLC_r D
F_RLC_r E
F_RLC_r H
F_RLC_r L
F_RLC_r A

FUNC RL_HL_
	PCONTB0 HL
	SETCX %d7
	roxl.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_RL_r l
FUNC RL_\l
	GET_\l %d0
	SETCX %d7
	roxl.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_RL_r B
F_RL_r C
F_RL_r D
F_RL_r E
F_RL_r H
F_RL_r L
F_RL_r A

FUNC RRC_HL_
	PCONTB0 HL
	ror.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_RRC_r l
FUNC RRC_\l
	GET_\l %d0
	ror.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_RRC_r B
F_RRC_r C
F_RRC_r D
F_RRC_r E
F_RRC_r H
F_RRC_r L
F_RRC_r A

FUNC RR_HL_
	PCONTB0 HL
	SETCX %d7
	roxr.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_RR_r l
FUNC RR_\l
	GET_\l %d0
	SETCX %d7
	roxr.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_RR_r B
F_RR_r C
F_RR_r D
F_RR_r E
F_RR_r H
F_RR_r L
F_RR_r A

FUNC SLA_HL_
	PCONTB0 HL
	lsl.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_SLA_r l
FUNC SLA_\l
	GET_\l %d0
	lsl.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_SLA_r B
F_SLA_r C
F_SLA_r D
F_SLA_r E
F_SLA_r H
F_SLA_r L
F_SLA_r A

FUNC SRA_HL_
	PCONTB0 HL
	asr.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_SRA_r l
FUNC SRA_\l
	GET_\l %d0
	asr.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_SRA_r B
F_SRA_r C
F_SRA_r D
F_SRA_r E
F_SRA_r H
F_SRA_r L
F_SRA_r A

FUNC SRL_HL_
	PCONTB0 HL
	lsr.b #1,%d0
	R_FLAGSC
	GET_HL %a0
	STORE
	RET 2,4

.macro F_SRL_r l
FUNC SRL_\l
	GET_\l %d0
	lsr.b #1,%d0
	R_FLAGSC
	SET_\l %d0
	RET 2,2
	.endm

F_SRL_r B
F_SRL_r C
F_SRL_r D
F_SRL_r E
F_SRL_r H
F_SRL_r L
F_SRL_r A

	.even
bitx0:
	FLAG1_Z
	FLAG1_P
	FLAG0_S
	bra bitxe
	.even
bitx1:
	FLAG0_Z
	FLAG0_P
	cmp.w #7,%a0
	bne bitx3
	FLAG1_S
	bra bitxe
	.even
bitx3:
	FLAG0_S
bitxe:
	FLAG1_H
	FLAG0_N
	RET 4,5

.macro F_BIT_b_IX_d n,m=_IX_d
FUNC BIT_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	move.w #\n,%a0
	btst.b #\n,%d0
	beq bitx0
	bra bitx1
	.endm

F_BIT_b_IX_d 0
F_BIT_b_IX_d 1
F_BIT_b_IX_d 2
F_BIT_b_IX_d 3
F_BIT_b_IX_d 4
F_BIT_b_IX_d 5
F_BIT_b_IX_d 6
F_BIT_b_IX_d 7

	.even
f_set_b_ix_d1:
	move.b %d0,%d7
	PCONTB PC,#2
	GET_IX %a0
	ext.w %d0
	add.w %d0,%a0
	move.b %d7,%d0
	STORE
	RET 4,5	

.macro F_SET_b_IX_d n,m=_IX_d
FUNC SET_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	bset.b #\n,%d0
	bra f_set_b_ix_d1
	.endm

F_SET_b_IX_d 0
F_SET_b_IX_d 1
F_SET_b_IX_d 2
F_SET_b_IX_d 3
F_SET_b_IX_d 4
F_SET_b_IX_d 5
F_SET_b_IX_d 6
F_SET_b_IX_d 7

.macro F_RES_b_IX_d n,m=_IX_d
FUNC RES_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	bclr.b #\n,%d0
	bra f_set_b_ix_d1
	.endm

F_RES_b_IX_d 0
F_RES_b_IX_d 1
F_RES_b_IX_d 2
F_RES_b_IX_d 3
F_RES_b_IX_d 4
F_RES_b_IX_d 5
F_RES_b_IX_d 6
F_RES_b_IX_d 7

FUNC RLC_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	rol.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC RL_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	SETCX %d7
	roxl.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC RRC_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	ror.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC RR_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	SETCX %d7
	roxr.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC SLA_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	lsl.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC SRA_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	asr.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

FUNC SRL_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	lsr.b #1,%d0
	R_FLAGSC
	bra f_set_b_ix_d1

.macro F_LD_r_IX_d l,m=_IX_d
FUNC LD_\l\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	SET_\l %d0
	RET 3,3
	.endm

F_LD_r_IX_d B
F_LD_r_IX_d C
F_LD_r_IX_d D
F_LD_r_IX_d E
F_LD_r_IX_d H
F_LD_r_IX_d L
F_LD_r_IX_d A

.macro F_LD_IX_d_r l
FUNC LD_IX_d_\l
	PCONTB PC,#2
	GET_IX %a0
	ext.w %d0
	add.w %d0,%a0
	GET_\l %d0
	STORE
	RET 3,5
	.endm

F_LD_IX_d_r B
F_LD_IX_d_r C
F_LD_IX_d_r D
F_LD_IX_d_r E
F_LD_IX_d_r H
F_LD_IX_d_r L
F_LD_IX_d_r A

FUNC LD_IX_d_n
	PCONTB PC,#2
	GET_IX %d7
	ext.w %d0
	add.w %d0,%d7
	PCONTB PC,#3
	move.w %d7,%a0
	STORE
	RET 4,5

FUNC LD_IX_nn
	PCONTW PC,#2
	SET_IX %d0
	RET 4,4

FUNC LD_IX_nn_
	PCONTW PC,#2
	PCONTWR0 %d0
	SET_IX %d0
	RET 4,5

FUNC LD_nn_IX
	PCONTW PC,#2
	move.w %d0,%a0
	GET_IX %d0
	STOREW
	RET 4,5

FUNC LD_SP_IX
	GET_IX %d0
	SET_SP %d0
	RET 2,3

FUNC PUSH_IX
	SUB_SP #2
	GET_SP %a0
	GET_IX %d0
	STOREW
	RET 2,4

FUNC POP_IX
	PCONTW0 SP
	SET_IX %d0
	ADD_SP #2
	RET 2,4

FUNC ADD_A_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 3,5

FUNC ADC_A_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 3,5

FUNC SUB_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 3,5

FUNC SBC_A_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 3,5

FUNC AND_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	exg %d0,%d7
	and.b %d7,%d0
	ANDFLAGS
	SET_A %d0
	RET 3,5

FUNC XOR_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	exg %d0,%d7
	eor.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 3,5

FUNC OR_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	exg %d0,%d7
	or.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 3,5

FUNC CP_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	RET 3,5

	.even
inc_ix_d1:
	move.b %d0,%d7
	PCONTB PC,#2
	GET_IX %a0
	ext.w %d0
	add.w %d0,%a0
	move.b %d7,%d0
	STORE
	RET 3,5

FUNC INC_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	move.b #1,%d7
	ADDFLAGS
	bra inc_ix_d1
	
FUNC DEC_IX_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IX,%d7
	move.b #1,%d7
	SUBFLAGS
	bra inc_ix_d1

	.even
f_add_ix_pp1:
	GET_IX %d7
	ADDHFLAGS
	SET_IX %d0
	RET 2,4

.macro F_ADD_IX_pp hl
FUNC ADD_IX_\hl
	GET_\hl %d0
	bra f_add_ix_pp1
	.endm

F_ADD_IX_pp BC
F_ADD_IX_pp DE
F_ADD_IX_pp IX
F_ADD_IX_pp SP

FUNC INC_IX
	ADD_IX #1
	RET 2,3

FUNC DEC_IX
	SUB_IX #1
	RET 2,3

FUNC JP_IX_
	GET_IX %d0
	SET_PC %d0
	RET 0,2

FUNC EX_SP_IX
	PCONTW0 SP
	move.w %d0,%d7
	GET_SP %a0
	GET_IX %d0
	STOREW
	SET_IX %d7
	RET 2,5

.macro F_BIT_b_IY_d n,m=_IY_d
FUNC BIT_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	move.w #\n,%a0
	btst.b #\n,%d0
	beq bitx0
	bra bitx1
	.endm

F_BIT_b_IY_d 0
F_BIT_b_IY_d 1
F_BIT_b_IY_d 2
F_BIT_b_IY_d 3
F_BIT_b_IY_d 4
F_BIT_b_IY_d 5
F_BIT_b_IY_d 6
F_BIT_b_IY_d 7

	.even
f_set_b_iy_d1:
	move.b %d0,%d7
	PCONTB PC,#2
	GET_IY %a0
	ext.w %d0
	add.w %d0,%a0
	move.b %d7,%d0
	STORE
	RET 4,5	


.macro F_SET_b_IY_d n,m=_IY_d
FUNC SET_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	bset.b #\n,%d0
	bra f_set_b_iy_d1
	.endm

F_SET_b_IY_d 0
F_SET_b_IY_d 1
F_SET_b_IY_d 2
F_SET_b_IY_d 3
F_SET_b_IY_d 4
F_SET_b_IY_d 5
F_SET_b_IY_d 6
F_SET_b_IY_d 7

.macro F_RES_b_IY_d n,m=_IY_d
FUNC RES_\n\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	bclr.b #\n,%d0
	bra f_set_b_iy_d1
	.endm

F_RES_b_IY_d 0
F_RES_b_IY_d 1
F_RES_b_IY_d 2
F_RES_b_IY_d 3
F_RES_b_IY_d 4
F_RES_b_IY_d 5
F_RES_b_IY_d 6
F_RES_b_IY_d 7

FUNC RLC_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	rol.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC RL_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	SETCX %d7
	roxl.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC RRC_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	ror.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC RR_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	SETCX %d7
	roxr.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC SLA_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	lsl.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC SRA_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	asr.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

FUNC SRL_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	lsr.b #1,%d0
	R_FLAGSC
	bra f_set_b_iy_d1

.macro F_LD_r_IY_d l,m=_IY_d
FUNC LD_\l\m
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	SET_\l %d0
	RET 3,3
	.endm

F_LD_r_IY_d B
F_LD_r_IY_d C
F_LD_r_IY_d D
F_LD_r_IY_d E
F_LD_r_IY_d H
F_LD_r_IY_d L
F_LD_r_IY_d A

.macro F_LD_IY_d_r l
FUNC LD_IY_d_\l
	PCONTB PC,#2
	GET_IY %a0
	ext.w %d0
	add.w %d0,%a0
	GET_\l %d0
	STORE
	RET 3,5
	.endm

F_LD_IY_d_r B
F_LD_IY_d_r C
F_LD_IY_d_r D
F_LD_IY_d_r E
F_LD_IY_d_r H
F_LD_IY_d_r L
F_LD_IY_d_r A

FUNC LD_IY_d_n
	PCONTB PC,#2
	GET_IY %d7
	ext.w %d0
	add.w %d0,%d7
	PCONTB PC,#3
	move.w %d7,%a0
	STORE
	RET 4,5

FUNC LD_IY_nn
	PCONTW PC,#2
	SET_IY %d0
	RET 4,4

FUNC LD_IY_nn_
	PCONTW PC,#2
	PCONTWR0 %d0
	SET_IY %d0
	RET 4,5

FUNC LD_nn_IY
	PCONTW PC,#2
	move.w %d0,%a0
	GET_IY %d0
	STOREW
	RET 4,5

FUNC LD_SP_IY
	GET_IY %d0
	SET_SP %d0
	RET 2,3

FUNC PUSH_IY
	SUB_SP #2
	GET_SP %a0
	GET_IY %d0
	STOREW
	RET 2,4

FUNC POP_IY
	PCONTW0 SP
	SET_IY %d0
	ADD_SP #2
	RET 2,4

FUNC ADD_A_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 3,5

FUNC ADC_A_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	ADDFLAGSC
	SET_A %d0
	RET 3,5

FUNC SUB_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 3,5

FUNC SBC_A_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_FLAGS %d7
	and.b #0x01,%d7
	add.b %d7,%d0
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	SET_A %d0
	RET 3,5

FUNC AND_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	exg %d0,%d7
	and.b %d7,%d0
	ANDFLAGS
	SET_A %d0
	RET 3,5

FUNC XOR_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	exg %d0,%d7
	eor.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 3,5

FUNC OR_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	exg %d0,%d7
	or.b %d7,%d0
	XORFLAGS
	SET_A %d0
	RET 3,5

FUNC CP_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	GET_A %d7
	exg %d7,%d0
	SUBFLAGSC
	RET 3,5

	.even
inc_iy_d1:
	move.b %d0,%d7
	PCONTB PC,#2
	GET_IY %a0
	ext.w %d0
	add.w %d0,%a0
	move.b %d7,%d0
	STORE
	RET 3,5

FUNC INC_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	move.b #1,%d7
	ADDFLAGS
	bra inc_iy_d1
	
FUNC DEC_IY_d
	PCONTB PC,#2
	move.b %d0,%d7
	PCONTB IY,%d7
	move.b #1,%d7
	SUBFLAGS
	bra inc_iy_d1

	.even
f_add_iy_pp1:
	GET_IY %d7
	ADDHFLAGS
	SET_IY %d0
	RET 2,4

.macro F_ADD_IY_pp hl
FUNC ADD_IY_\hl
	GET_\hl %d0
	bra f_add_iy_pp1
	.endm

F_ADD_IY_pp BC
F_ADD_IY_pp DE
F_ADD_IY_pp IY
F_ADD_IY_pp SP

FUNC INC_IY
	ADD_IY #1
	RET 2,3

FUNC DEC_IY
	SUB_IY #1
	RET 2,3

FUNC JP_IY_
	GET_IY %d0
	SET_PC %d0
	RET 0,2

FUNC EX_SP_IY
	PCONTW0 SP
	move.w %d0,%d7
	GET_SP %a0
	GET_IY %d0
	STOREW
	SET_IY %d7
	RET 2,5

# Undocumented

FUNC LD_IXh_n
  PCONTB PC,#2
  SET_IX_I %d0
  RET 3,5

FUNC LD_IXl_n
  PCONTB PC,#2
  SET_IX_X %d0
  RET 3,5

FUNC LD_IXh_A
  GET_A %d0
  SET_IX_I %d0
  RET 2,5

FUNC LD_IXl_A
  GET_A %d0
  SET_IX_X %d0
  RET 2,5

FUNC LD_IYh_n
  PCONTB PC,#2
  SET_IY_I %d0
  RET 3,5

FUNC LD_IYl_n
  PCONTB PC,#2
  SET_IY_Y %d0
  RET 3,5

FUNC LD_IYh_A
  GET_A %d0
  SET_IY_I %d0
  RET 2,5

FUNC LD_IYl_A
  GET_A %d0
  SET_IY_Y %d0
  RET 2,5


.even
.comm opcode,1024
.comm opcodeed,1024
.comm opcodecb,1024
.comm opcodedd,1024
.comm opcodefd,1024
.comm opcodeddcb,1024
.comm opcodefdcb,1024
.comm REG_A,1
.comm REG_F,1
.comm REG_B,1
.comm REG_C,1
.comm REG_D,1
.comm REG_E,1
.comm REG_H,1
.comm REG_L,1
.comm REG_AF_,2
.comm REG_BC_,2
.comm REG_DE_,2
.comm REG_HL_,2
.comm REG_IX,2
.comm REG_IY,2
.comm REG_PC,2
.comm REG_SP,2
.comm REG_I,1
.comm REG_R,1
.comm REG_FLAGS,1
.comm REG_IM,1
.comm REG_IFF1,1
.comm REG_IFF2,1
.comm REG_NEI,1
.comm REG_flash,1
.comm ROM,4
.comm MEM,4
.comm TIME,4
.comm time,4
.comm TIME_INT,4
.comm TIME_FLASH,4
.comm KEYBOARD0,1
.comm KEYBOARD1,1
.comm KEYBOARD2,1
.comm KEYBOARD3,1
.comm KEYBOARD4,1
.comm KEYBOARD5,1
.comm KEYBOARD6,1
.comm KEYBOARD7,1
.comm KEMPSTON,1
.comm sc,4
.comm vx,2
.comm vy,2
.comm trace,1
.comm temp,4
.comm SNA,27
.comm rsc,1
.comm sb,0x1b00
.comm memrom,4
.comm memram,4
